"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ByteParser = exports.ByteReader = exports.ParseResult = exports.finishedParser = exports.ParseStep = void 0;
const base_1 = require("./base");
const recurse = true;
const dontRecurse = false;
const needMoreData = new Error();
class ParseStep {
    onTruncation(reader) {
        throw Error(reader.remainingSize() + ' bytes remain after finished parsing');
    }
}
exports.ParseStep = ParseStep;
class FinishedParser extends ParseStep {
    parse() {
        throw Error('No initial parser installed: you must use startWith(...)');
    }
}
exports.finishedParser = new FinishedParser();
class ParseResult {
    constructor(result, nextStep) {
        this.result = result;
        this.nextStep = nextStep;
    }
}
exports.ParseResult = ParseResult;
class ByteReader {
    constructor(input) {
        this.input = base_1.emptyBuffer;
        this.off = 0;
        this.setInput(input);
    }
    setInput(input) {
        this.input = input;
        this.off = 0;
    }
    hasRemaining() {
        return this.off < this.input.length;
    }
    remainingSize() {
        return this.input.length - this.off;
    }
    remainingData() {
        return this.hasRemaining() ? this.input.slice(this.off) : base_1.emptyBuffer;
    }
    ensure(n) {
        if (this.remainingSize() < n) {
            throw needMoreData;
        }
    }
    take(n) {
        if (this.off + n <= this.input.length) {
            const o = this.off;
            this.off = o + n;
            return this.input.slice(o, this.off);
        }
        else {
            throw needMoreData;
        }
    }
}
exports.ByteReader = ByteReader;
class ByteParser {
    constructor(out) {
        this.out = out;
        this.current = exports.finishedParser;
        this.isCompleted = false;
        this.hasData = false;
        this.reader = new ByteReader(base_1.emptyBuffer);
        this.buffer = base_1.emptyBuffer;
    }
    parse(chunk) {
        this.buffer = (0, base_1.concat)(this.buffer, chunk);
        this.hasData = chunk.length > 0;
        while (this.hasData && !this.isCompleted) {
            this.doParse(1000);
        }
    }
    flush() {
        if (!this.isCompleted) {
            if (this.buffer.length > 0) {
                try {
                    this.reader.setInput(this.buffer);
                    this.current.onTruncation(this.reader);
                    this.complete();
                }
                catch (error) {
                    this.fail(error);
                }
            }
            else {
                this.complete();
            }
        }
    }
    startWith(step) {
        this.current = step;
    }
    complete() {
        this.isCompleted = true;
        this.buffer = base_1.emptyBuffer;
        this.reader = null;
        this.out.complete();
    }
    fail(error) {
        error.message = 'Parsing failed: ' + (error && error.message ? error.message : '');
        this.isCompleted = true;
        this.buffer = base_1.emptyBuffer;
        this.reader = null;
        this.out.fail(error);
    }
    doParseInner() {
        if (this.buffer.length > 0) {
            this.reader.setInput(this.buffer);
            try {
                const parseResult = this.current.parse(this.reader);
                if (parseResult.result) {
                    this.out.next(parseResult.result);
                }
                if (parseResult.nextStep === exports.finishedParser) {
                    this.complete();
                    return dontRecurse;
                }
                else {
                    this.buffer = this.reader.remainingData();
                    this.current = parseResult.nextStep;
                    if (!this.reader.hasRemaining()) {
                        this.hasData = false;
                    }
                    // If this step didn't produce a result, continue parsing.
                    if (!parseResult.result) {
                        return recurse;
                    }
                    else {
                        return dontRecurse;
                    }
                }
            }
            catch (error) {
                if (error === needMoreData) {
                    this.hasData = false;
                    return dontRecurse;
                }
                this.fail(error);
                return dontRecurse;
            }
        }
        else {
            this.hasData = false;
            return dontRecurse;
        }
    }
    doParse(remainingRecursions) {
        if (remainingRecursions === 0) {
            this.fail(new Error("Parsing logic didn't produce result. Aborting processing to avoid infinite cycles."));
        }
        else {
            const doRecurse = this.doParseInner();
            if (doRecurse) {
                this.doParse(remainingRecursions - 1);
            }
        }
    }
}
exports.ByteParser = ByteParser;
//# sourceMappingURL=byte-parser.js.map