"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Fragments = exports.Fragment = exports.Item = exports.Sequence = exports.SequenceDelimitationElement = exports.ItemDelimitationElement = exports.FragmentElement = exports.ItemElement = exports.FragmentsElement = exports.SequenceElement = exports.ValueElement = exports.preambleElement = exports.UnknownElement = exports.ElementSet = exports.Element = void 0;
const base_1 = require("./base");
const lookup_1 = require("./lookup");
const dicom_parts_1 = require("./dicom-parts");
const tag_1 = require("./tag");
const value_1 = require("./value");
const vr_1 = require("./vr");
class Element {
    constructor(bigEndian = false) {
        this.bigEndian = bigEndian;
    }
    toBytes() {
        return base_1.emptyBuffer;
    }
    toParts() {
        return [];
    }
}
exports.Element = Element;
class ElementSet {
    constructor(tag, vr, bigEndian = false, explicitVR = true) {
        this.tag = tag;
        this.vr = vr;
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
    }
    toBytes() {
        return base_1.emptyBuffer;
    }
    toElements() {
        return [];
    }
}
exports.ElementSet = ElementSet;
class UnknownElement extends Element {
    constructor(bigEndian) {
        super(bigEndian);
    }
}
exports.UnknownElement = UnknownElement;
class PreambleElement extends Element {
    constructor() {
        super(false);
    }
    toBytes() {
        return (0, base_1.concat)(Buffer.from(new Array(128).fill(0)), Buffer.from('DICM'));
    }
    toString() {
        return 'PreambleElement(0, ..., 0, D, I, C, M)';
    }
    toParts() {
        return [new dicom_parts_1.PreamblePart(this.toBytes())];
    }
}
exports.preambleElement = new PreambleElement();
class ValueElement extends ElementSet {
    constructor(tag, vr, value, bigEndian, explicitVR) {
        super(tag, vr, bigEndian, explicitVR);
        this.value = value;
        this.length = value.length;
    }
    setValue(value) {
        return new ValueElement(this.tag, this.vr, value.ensurePadding(this.vr), this.bigEndian, this.explicitVR);
    }
    toBytes() {
        return this.toParts()
            .map((p) => p.bytes)
            .reduce(base_1.concat);
    }
    toParts() {
        const headerPart = dicom_parts_1.HeaderPart.create(this.tag, this.vr, this.length, this.bigEndian, this.explicitVR);
        if (this.length > 0) {
            return [headerPart, new dicom_parts_1.ValueChunk(this.bigEndian, this.value.bytes, true)];
        }
        else {
            return [headerPart];
        }
    }
    toElements() {
        return [this];
    }
    toString() {
        const strings = this.value.toStrings(this.vr, this.bigEndian, base_1.defaultCharacterSet);
        const s = strings.join(base_1.multiValueDelimiter);
        const vm = strings.length + '';
        return ('ValueElement(' +
            (0, base_1.tagToString)(this.tag) +
            ' ' +
            this.vr.name +
            ' [' +
            s +
            '] # ' +
            this.length +
            ', ' +
            vm +
            ' ' +
            lookup_1.Lookup.keywordOf(this.tag) || '' + ')');
    }
}
exports.ValueElement = ValueElement;
class SequenceElement extends Element {
    constructor(tag, length = base_1.indeterminateLength, bigEndian, explicitVR = true) {
        super(bigEndian);
        this.tag = tag;
        this.length = length;
        this.explicitVR = explicitVR;
        this.tag = tag;
        this.indeterminate = this.length === base_1.indeterminateLength;
    }
    toBytes() {
        return dicom_parts_1.HeaderPart.create(this.tag, vr_1.VR.SQ, this.length, this.bigEndian, this.explicitVR).bytes;
    }
    toParts() {
        return [new dicom_parts_1.SequencePart(this.tag, this.length, this.bigEndian, this.explicitVR, this.toBytes())];
    }
    toString() {
        return ('SequenceElement(' + (0, base_1.tagToString)(this.tag) + ' SQ # ' + this.length + ' ' + lookup_1.Lookup.keywordOf(this.tag) ||
            '' + ')');
    }
}
exports.SequenceElement = SequenceElement;
class FragmentsElement extends Element {
    constructor(tag, vr, bigEndian, explicitVR = true) {
        super(bigEndian);
        this.tag = tag;
        this.vr = vr;
        this.explicitVR = explicitVR;
    }
    toBytes() {
        return this.toParts()[0].bytes;
    }
    toParts() {
        return [
            new dicom_parts_1.FragmentsPart(this.tag, base_1.indeterminateLength, this.vr, this.bigEndian, this.explicitVR, dicom_parts_1.HeaderPart.create(this.tag, this.vr, base_1.indeterminateLength, this.bigEndian, this.explicitVR).bytes),
        ];
    }
    toString() {
        return ('FragmentsElement(' + (0, base_1.tagToString)(this.tag) + ' ' + this.vr.name + ' # ' + lookup_1.Lookup.keywordOf(this.tag) ||
            '' + ')');
    }
}
exports.FragmentsElement = FragmentsElement;
class ItemElement extends Element {
    constructor(length = base_1.indeterminateLength, bigEndian) {
        super(bigEndian);
        this.length = length;
        this.indeterminate = this.length === base_1.indeterminateLength;
    }
    toBytes() {
        return (0, base_1.concat)((0, base_1.tagToBytes)(tag_1.Tag.Item, this.bigEndian), (0, base_1.intToBytes)(this.length, this.bigEndian));
    }
    toParts() {
        return [new dicom_parts_1.ItemPart(this.length, this.bigEndian, this.toBytes())];
    }
    toString() {
        return 'ItemElement(length = ' + this.length + ')';
    }
}
exports.ItemElement = ItemElement;
class FragmentElement extends Element {
    constructor(length, value, bigEndian) {
        super(bigEndian);
        this.length = length;
        this.value = value;
    }
    toBytes() {
        return this.toParts()
            .map((p) => p.bytes)
            .reduce(base_1.concat);
    }
    toParts() {
        const itemParts = new ItemElement(this.value.length, this.bigEndian).toParts();
        if (this.value.length !== 0) {
            itemParts.push(new dicom_parts_1.ValueChunk(this.bigEndian, this.value.bytes, true));
        }
        return itemParts;
    }
    toString() {
        return 'FragmentElement(length = ' + this.length + ')';
    }
}
exports.FragmentElement = FragmentElement;
class ItemDelimitationElement extends Element {
    constructor(bigEndian) {
        super(bigEndian);
    }
    toBytes() {
        return (0, base_1.concat)((0, base_1.tagToBytes)(tag_1.Tag.ItemDelimitationItem, this.bigEndian), Buffer.from([0, 0, 0, 0]));
    }
    toParts() {
        return [new dicom_parts_1.ItemDelimitationPart(this.bigEndian, this.toBytes())];
    }
    toString() {
        return 'ItemDelimitationElement';
    }
}
exports.ItemDelimitationElement = ItemDelimitationElement;
class SequenceDelimitationElement extends Element {
    constructor(bigEndian) {
        super(bigEndian);
    }
    toBytes() {
        return (0, base_1.concat)((0, base_1.tagToBytes)(tag_1.Tag.SequenceDelimitationItem, this.bigEndian), Buffer.from([0, 0, 0, 0]));
    }
    toParts() {
        return [new dicom_parts_1.SequenceDelimitationPart(this.bigEndian, this.toBytes())];
    }
    toString() {
        return 'SequenceDelimitationElement';
    }
}
exports.SequenceDelimitationElement = SequenceDelimitationElement;
class Sequence extends ElementSet {
    constructor(tag, length = base_1.indeterminateLength, items = [], bigEndian, explicitVR) {
        super(tag, vr_1.VR.SQ, bigEndian, explicitVR);
        this.tag = tag;
        this.length = length;
        this.items = items;
        this.indeterminate = length === base_1.indeterminateLength;
        this.size = items.length;
    }
    item(index) {
        return this.items.length >= index ? this.items[index - 1] : undefined;
    }
    addItem(item) {
        const newItems = (0, base_1.appendToArray)(item, this.items);
        const newLength = this.indeterminate ? this.length : this.length + item.toBytes().length;
        return new Sequence(this.tag, newLength, newItems, this.bigEndian, this.explicitVR);
    }
    removeItem(index) {
        const newItems = this.items.slice();
        newItems.splice(index - 1, 1);
        const newLength = this.indeterminate ? this.length : this.length - this.item(index).toBytes().length;
        return new Sequence(this.tag, newLength, newItems, this.bigEndian, this.explicitVR);
    }
    toBytes() {
        return this.toElements()
            .map((e) => e.toBytes())
            .reduce(base_1.concat, base_1.emptyBuffer);
    }
    toElements() {
        const elements = [];
        elements.push(new SequenceElement(this.tag, this.length, this.bigEndian, this.explicitVR));
        for (let i = 1; i <= this.items.length; i++) {
            const itemElements = this.item(i).toElements();
            itemElements.forEach((e) => elements.push(e));
        }
        if (this.indeterminate) {
            elements.push(new SequenceDelimitationElement(this.bigEndian));
        }
        return elements;
    }
    setItem(index, item) {
        const newItems = this.items.slice();
        newItems[index - 1] = item;
        return new Sequence(this.tag, this.length, newItems, this.bigEndian, this.explicitVR);
    }
    toString() {
        return ('Sequence(' +
            (0, base_1.tagToString)(this.tag) +
            ' SQ # ' +
            this.length +
            ' ' +
            this.size +
            ' ' +
            lookup_1.Lookup.keywordOf(this.tag) || '' + ')');
    }
}
exports.Sequence = Sequence;
class Item {
    constructor(elements, length = base_1.indeterminateLength, bigEndian = false) {
        this.elements = elements;
        this.length = length;
        this.bigEndian = bigEndian;
        this.indeterminate = length === base_1.indeterminateLength;
    }
    toElements() {
        const elements = [];
        elements.push(new ItemElement(this.length, this.bigEndian));
        this.elements.toElements(false).forEach((e) => elements.push(e));
        if (this.indeterminate) {
            elements.push(new ItemDelimitationElement(this.bigEndian));
        }
        return elements;
    }
    toBytes() {
        return this.toElements()
            .map((e) => e.toBytes())
            .reduce(base_1.concat);
    }
    setElements(elements) {
        const newLength = this.indeterminate ? base_1.indeterminateLength : elements.toBytes(false).length;
        return new Item(elements, newLength, this.bigEndian);
    }
    toString() {
        return 'Item(length = ' + this.length + ', elements size = ' + this.elements.size + ')';
    }
}
exports.Item = Item;
class Fragment {
    constructor(length, value, bigEndian = false) {
        this.length = length;
        this.value = value;
        this.bigEndian = bigEndian;
    }
    toElement() {
        return new FragmentElement(this.length, this.value, this.bigEndian);
    }
    toString() {
        return 'Fragment(length = ' + this.length + ', value length = ' + this.value.length + ')';
    }
}
exports.Fragment = Fragment;
class Fragments extends ElementSet {
    constructor(tag, vr, offsets, fragments = [], bigEndian, explicitVR) {
        super(tag, vr, bigEndian, explicitVR);
        this.tag = tag;
        this.vr = vr;
        this.offsets = offsets;
        this.fragments = fragments;
        this.size = fragments.length;
    }
    fragment(index) {
        return this.fragments.length > index ? undefined : this.fragments[index - 1];
    }
    frameCount() {
        return this.offsets === undefined && this.fragments.length === 0
            ? 0
            : this.offsets === undefined
                ? 1
                : this.offsets.length;
    }
    addFragment(fragment) {
        if (this.size === 0 && this.offsets === undefined) {
            const bytes = fragment.value.bytes;
            const offsets = [];
            for (let i = 0; i < bytes.length; i += 4) {
                offsets.push((0, base_1.bytesToUInt)(bytes.slice(i), fragment.bigEndian));
            }
            return new Fragments(this.tag, this.vr, offsets, this.fragments, this.bigEndian, this.explicitVR);
        }
        else {
            return new Fragments(this.tag, this.vr, this.offsets, (0, base_1.appendToArray)(fragment, this.fragments), this.bigEndian, this.explicitVR);
        }
    }
    toBytes() {
        return this.toElements()
            .map((e) => e.toBytes())
            .reduce(base_1.concat);
    }
    toElements() {
        const elements = [];
        elements.push(new FragmentsElement(this.tag, this.vr, this.bigEndian, this.explicitVR));
        if (this.offsets !== undefined) {
            elements.push(new FragmentElement(4 * this.offsets.length, new value_1.Value(this.offsets
                .map((offset) => (0, base_1.intToBytes)(offset, this.bigEndian), this.bigEndian)
                .reduce(base_1.concat, base_1.emptyBuffer)), this.bigEndian));
        }
        else {
            elements.push(new FragmentElement(0, value_1.Value.empty()));
        }
        for (let i = 1; i <= this.fragments.length; i++) {
            elements.push(this.fragment(i).toElement());
        }
        elements.push(new SequenceDelimitationElement(this.bigEndian));
        return elements;
    }
    setFragment(index, fragment) {
        const newFragments = this.fragments.slice();
        newFragments[index - 1] = fragment;
        return new Fragments(this.tag, this.vr, this.offsets, newFragments, this.bigEndian, this.explicitVR);
    }
    toString() {
        return `Fragments(${(0, base_1.tagToString)(this.tag)} ${this.vr.name} # ${this.fragments.length} ${lookup_1.Lookup.keywordOf(this.tag) || ''})`;
    }
}
exports.Fragments = Fragments;
//# sourceMappingURL=dicom-elements.js.map