"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.deflateDatasetFlow = exports.toUtf8Flow = exports.toIndeterminateLengthSequences = exports.fmiGroupLengthFlow = exports.validateContextFlow = exports.ValidationContext = exports.headerFilter = exports.fmiDiscardFilter = exports.groupLengthDiscardFilter = exports.denyFilter = exports.allowFilter = exports.tagFilter = exports.stopTagFlow = exports.toBytesFlow = void 0;
const stream_1 = require("stream");
const zlib_1 = __importDefault(require("zlib"));
const base_1 = require("./base");
const character_sets_1 = require("./character-sets");
const collect_flow_1 = require("./collect-flow");
const detour_1 = require("./detour");
const dicom_flow_1 = require("./dicom-flow");
const modify_flow_1 = require("./modify-flow");
const dicom_parts_1 = require("./dicom-parts");
const tag_1 = require("./tag");
const tag_path_1 = require("./tag-path");
const tag_path_2 = require("./tag-path");
const tag_tree_1 = require("./tag-tree");
const uid_1 = require("./uid");
const vr_1 = require("./vr");
function toBytesFlow() {
    return new stream_1.Transform({
        writableObjectMode: true,
        transform(chunk, encoding, callback) {
            this.push(chunk.bytes);
            process.nextTick(() => callback());
        },
    });
}
exports.toBytesFlow = toBytesFlow;
function stopTagFlow(tag) {
    let endReached = false;
    return (0, base_1.pipe)((0, dicom_flow_1.createFlow)(new (class extends (0, dicom_flow_1.InSequence)((0, dicom_flow_1.GuaranteedDelimitationEvents)((0, dicom_flow_1.InFragments)(dicom_flow_1.IdentityFlow))) {
        onHeader(part) {
            const out = super.onHeader(part);
            return this.inSequence() || part.tag < tag ? out : [dicom_flow_1.dicomEndMarker];
        }
    })()), new stream_1.Transform({
        objectMode: true,
        transform(chunk, encoding, callback) {
            if (!endReached) {
                if (chunk === dicom_flow_1.dicomEndMarker) {
                    endReached = true;
                    this.push(null);
                }
                else {
                    this.push(chunk);
                }
            }
            process.nextTick(() => callback());
        },
    }));
}
exports.stopTagFlow = stopTagFlow;
function tagFilter(keepCondition, defaultCondition = () => true, logGroupLengthWarnings = false) {
    return (0, dicom_flow_1.createFlow)(new (class extends (0, dicom_flow_1.TagPathTracking)((0, dicom_flow_1.GuaranteedDelimitationEvents)((0, dicom_flow_1.GuaranteedValueEvent)((0, dicom_flow_1.GroupLengthWarnings)((0, dicom_flow_1.InFragments)(dicom_flow_1.DeferToPartFlow))))) {
        constructor() {
            super();
            this.keeping = false;
            this.setSilent(!logGroupLengthWarnings);
        }
        onPart(part) {
            this.keeping = this.tagPath === tag_path_2.emptyTagPath ? defaultCondition(part) : keepCondition(this.tagPath);
            return this.keeping ? [part] : [];
        }
    })());
}
exports.tagFilter = tagFilter;
function allowFilter(allowlist, defaultCondition, logGroupLengthWarnings) {
    return tagFilter((currentPath) => allowlist.some((t) => t.hasTrunk(currentPath) || t.isTrunkOf(currentPath)), defaultCondition, logGroupLengthWarnings);
}
exports.allowFilter = allowFilter;
function denyFilter(denylist, defaultCondition, logGroupLengthWarnings) {
    return tagFilter((currentPath) => !denylist.some((t) => t.isTrunkOf(currentPath)), defaultCondition, logGroupLengthWarnings);
}
exports.denyFilter = denyFilter;
function groupLengthDiscardFilter() {
    return tagFilter((tagPath) => !(0, base_1.isGroupLength)(tagPath.tag()) || (0, base_1.isFileMetaInformation)(tagPath.tag()));
}
exports.groupLengthDiscardFilter = groupLengthDiscardFilter;
function fmiDiscardFilter() {
    return tagFilter((tagPath) => !(0, base_1.isFileMetaInformation)(tagPath.tag()), () => false);
}
exports.fmiDiscardFilter = fmiDiscardFilter;
function headerFilter(keepCondition, logGroupLengthWarnings = false) {
    return (0, dicom_flow_1.createFlow)(new (class extends (0, dicom_flow_1.GroupLengthWarnings)((0, dicom_flow_1.InFragments)(dicom_flow_1.DeferToPartFlow)) {
        constructor() {
            super();
            this.keeping = false;
            this.setSilent(!logGroupLengthWarnings);
        }
        onPart(part) {
            if (part instanceof dicom_parts_1.HeaderPart) {
                this.keeping = keepCondition(part);
                return this.keeping ? [part] : [];
            }
            if (part instanceof dicom_parts_1.ValueChunk) {
                return this.keeping ? [part] : [];
            }
            this.keeping = true;
            return [part];
        }
    })());
}
exports.headerFilter = headerFilter;
class ValidationContext {
    constructor(sopClassUID, transferSyntaxUID) {
        this.sopClassUID = sopClassUID;
        this.transferSyntaxUID = transferSyntaxUID;
    }
}
exports.ValidationContext = ValidationContext;
function validateContextFlow(contexts) {
    return (0, base_1.pipe)((0, collect_flow_1.collectFromTagPathsFlow)([
        tag_tree_1.TagTree.fromTag(tag_1.Tag.MediaStorageSOPClassUID),
        tag_tree_1.TagTree.fromTag(tag_1.Tag.TransferSyntaxUID),
        tag_tree_1.TagTree.fromTag(tag_1.Tag.SOPClassUID),
    ], 'validatecontext'), (0, dicom_flow_1.createFlow)(new (class extends dicom_flow_1.DeferToPartFlow {
        onPart(part) {
            if (part instanceof dicom_parts_1.ElementsPart && part.label === 'validatecontext') {
                let scuid = part.elements.stringByTag(tag_1.Tag.MediaStorageSOPClassUID);
                if (scuid === undefined) {
                    scuid = part.elements.stringByTag(tag_1.Tag.SOPClassUID);
                }
                if (scuid === undefined) {
                    scuid = '<empty>';
                }
                let tsuid = part.elements.stringByTag(tag_1.Tag.TransferSyntaxUID);
                if (tsuid === undefined) {
                    tsuid = '<empty>';
                }
                if (contexts.findIndex((c) => c.sopClassUID === scuid && c.transferSyntaxUID === tsuid) >= 0) {
                    return [];
                }
                else {
                    throw Error('The presentation context [SOPClassUID = ' +
                        scuid +
                        ', TransferSyntaxUID = ' +
                        tsuid +
                        '] is not supported');
                }
            }
            return [part];
        }
    })()));
}
exports.validateContextFlow = validateContextFlow;
function fmiGroupLengthFlow() {
    return (0, base_1.pipe)((0, collect_flow_1.collectFlow)((tagPath) => tagPath.isRoot() && (0, base_1.isFileMetaInformation)(tagPath.tag()), (tagPath) => !(0, base_1.isFileMetaInformation)(tagPath.tag()), 'fmigrouplength'), tagFilter((tagPath) => !(0, base_1.isFileMetaInformation)(tagPath.tag()), () => true, false), (0, dicom_flow_1.createFlow)(new (class extends (0, dicom_flow_1.EndEvent)(dicom_flow_1.DeferToPartFlow) {
        constructor() {
            super(...arguments);
            this.fmi = [];
            this.hasEmitted = false;
        }
        onEnd() {
            return this.hasEmitted ? [] : this.fmi;
        }
        onPart(part) {
            if (part instanceof dicom_parts_1.ElementsPart && part.label === 'fmigrouplength') {
                const elements = part.elements;
                if (elements.data.length > 0) {
                    const bigEndian = elements.data[0].bigEndian;
                    const explicitVR = elements.data[0].explicitVR;
                    const fmiElementsNoLength = elements.filter((e) => e.tag !== tag_1.Tag.FileMetaInformationGroupLength);
                    const length = fmiElementsNoLength.data
                        .map((e) => e.toBytes().length)
                        .reduce((l1, l2) => l1 + l2, 0);
                    const lengthHeader = dicom_parts_1.HeaderPart.create(tag_1.Tag.FileMetaInformationGroupLength, vr_1.VR.UL, 4, bigEndian, explicitVR);
                    const lengthChunk = new dicom_parts_1.ValueChunk(bigEndian, (0, base_1.intToBytes)(length, bigEndian), true);
                    this.fmi = (0, base_1.concatArrays)([lengthHeader, lengthChunk], fmiElementsNoLength.toParts(false));
                }
                return [];
            }
            if (!this.hasEmitted && part.bytes.length > 0) {
                this.hasEmitted = true;
                return part instanceof dicom_parts_1.PreamblePart
                    ? (0, base_1.prependToArray)(part, this.fmi)
                    : (0, base_1.appendToArray)(part, this.fmi);
            }
            return [part];
        }
    })()));
}
exports.fmiGroupLengthFlow = fmiGroupLengthFlow;
function toIndeterminateLengthSequences() {
    return (0, dicom_flow_1.createFlow)(new (class extends (0, dicom_flow_1.GuaranteedDelimitationEvents)((0, dicom_flow_1.InFragments)(dicom_flow_1.IdentityFlow)) {
        constructor() {
            super(...arguments);
            this.indeterminateBytes = Buffer.from([0xff, 0xff, 0xff, 0xff]);
        }
        onSequence(part) {
            return super.onSequence(part).map((p) => {
                if (p instanceof dicom_parts_1.SequencePart && !p.indeterminate) {
                    return new dicom_parts_1.SequencePart(part.tag, base_1.indeterminateLength, part.bigEndian, part.explicitVR, (0, base_1.concat)(part.bytes.slice(0, part.bytes.length - 4), this.indeterminateBytes));
                }
                return p;
            });
        }
        onSequenceDelimitation(part) {
            const out = super.onSequenceDelimitation(part);
            if (part.bytes.length <= 0) {
                out.push(new dicom_parts_1.SequenceDelimitationPart(part.bigEndian, (0, base_1.sequenceDelimitation)(part.bigEndian)));
            }
            return out;
        }
        onItem(part) {
            return super.onItem(part).map((p) => {
                if (p instanceof dicom_parts_1.ItemPart && !this.inFragments && !p.indeterminate) {
                    return new dicom_parts_1.ItemPart(base_1.indeterminateLength, part.bigEndian, (0, base_1.concat)(part.bytes.slice(0, part.bytes.length - 4), this.indeterminateBytes));
                }
                return p;
            });
        }
        onItemDelimitation(part) {
            const out = super.onItemDelimitation(part);
            if (part.bytes.length <= 0) {
                out.push(new dicom_parts_1.ItemDelimitationPart(part.bigEndian, (0, base_1.itemDelimitation)(part.bigEndian)));
            }
            return out;
        }
    })());
}
exports.toIndeterminateLengthSequences = toIndeterminateLengthSequences;
function toUtf8Flow() {
    return (0, base_1.pipe)((0, collect_flow_1.collectFromTagPathsFlow)([tag_tree_1.TagTree.fromTag(tag_1.Tag.SpecificCharacterSet)], 'toutf8'), (0, modify_flow_1.modifyFlow)([], [new modify_flow_1.TagInsertion(tag_path_1.TagPath.fromTag(tag_1.Tag.SpecificCharacterSet), () => Buffer.from('ISO_IR 192'))]), (0, dicom_flow_1.createFlow)(new (class extends (0, dicom_flow_1.GroupLengthWarnings)((0, dicom_flow_1.InFragments)(dicom_flow_1.IdentityFlow)) {
        constructor() {
            super(...arguments);
            this.characterSets = character_sets_1.defaultCharacterSet;
            this.currentValue = base_1.emptyBuffer;
        }
        onHeader(part) {
            if (part.length > 0 && character_sets_1.CharacterSets.isVrAffectedBySpecificCharacterSet(part.vr)) {
                this.currentHeader = part;
                this.currentValue = base_1.emptyBuffer;
                return [];
            }
            else {
                this.currentHeader = undefined;
                return [part];
            }
        }
        onValueChunk(part) {
            if (this.currentHeader !== undefined && !this.inFragments) {
                this.currentValue = (0, base_1.concat)(this.currentValue, part.bytes);
                if (part.last) {
                    const header = this.currentHeader;
                    this.currentHeader = undefined;
                    const newValue = Buffer.from(this.characterSets.decode(this.currentValue, header.vr));
                    const newLength = newValue.length;
                    return [
                        header.withUpdatedLength(newLength),
                        new dicom_parts_1.ValueChunk(header.bigEndian, newValue, true),
                    ];
                }
                else {
                    return [];
                }
            }
            else {
                return [part];
            }
        }
        onPart(part) {
            if (part instanceof dicom_parts_1.ElementsPart && part.label === 'toutf8') {
                if (part.elements.contains(tag_1.Tag.SpecificCharacterSet)) {
                    this.characterSets = character_sets_1.CharacterSets.fromBytes(part.elements.bytesByTag(tag_1.Tag.SpecificCharacterSet));
                }
                return [];
            }
            return [part];
        }
    })()));
}
exports.toUtf8Flow = toUtf8Flow;
class DeflateDatasetFlow extends detour_1.Detour {
    constructor() {
        super({ objectMode: true });
        this.collectingTs = false;
        this.tsBytes = base_1.emptyBuffer;
    }
    process(part) {
        if (part instanceof dicom_parts_1.HeaderPart) {
            if (part.isFmi) {
                this.collectingTs = part.tag === tag_1.Tag.TransferSyntaxUID;
                this.push(part);
            }
            else {
                if (this.tsBytes.toString().trim() === uid_1.UID.DeflatedExplicitVRLittleEndian) {
                    const toDeflatedChunk = new stream_1.Transform({
                        readableObjectMode: true,
                        transform(chunk, encoding, cb) {
                            this.push(new dicom_parts_1.DeflatedChunk(false, chunk));
                            process.nextTick(() => cb());
                        },
                    });
                    this.setDetourFlow((0, base_1.pipe)(toBytesFlow(), zlib_1.default.createDeflateRaw(), toDeflatedChunk));
                    this.setDetour(true, part);
                }
                else {
                    this.push(part);
                }
            }
        }
        else if (part instanceof dicom_parts_1.ValueChunk && this.collectingTs) {
            this.tsBytes = (0, base_1.concat)(this.tsBytes, part.bytes);
            this.push(part);
        }
        else {
            this.push(part);
        }
    }
}
function deflateDatasetFlow() {
    return new DeflateDatasetFlow();
}
exports.deflateDatasetFlow = deflateDatasetFlow;
//# sourceMappingURL=dicom-flows.js.map