"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ElementsPart = exports.UnknownPart = exports.FragmentsPart = exports.SequenceDelimitationPart = exports.SequencePart = exports.ItemDelimitationPart = exports.ItemPart = exports.DeflatedChunk = exports.ValueChunk = exports.HeaderPart = exports.PreamblePart = exports.MetaPart = exports.DicomPart = void 0;
const base_1 = require("./base");
class DicomPart {
    constructor(bigEndian, bytes) {
        this.bigEndian = bigEndian;
        this.bytes = bytes;
    }
}
exports.DicomPart = DicomPart;
class MetaPart extends DicomPart {
    constructor() {
        super(false, base_1.emptyBuffer);
    }
}
exports.MetaPart = MetaPart;
class PreamblePart extends DicomPart {
    constructor(bytes) {
        super(false, bytes);
    }
    toString() {
        return 'Preamble []';
    }
}
exports.PreamblePart = PreamblePart;
class HeaderPart extends DicomPart {
    static create(tag, vr, length, bigEndian = false, explicitVR = true) {
        const bytes = explicitVR
            ? vr.headerLength === 8
                ? Buffer.concat([(0, base_1.tagToBytes)(tag, bigEndian), Buffer.from(vr.name), (0, base_1.shortToBytes)(length, bigEndian)], 8)
                : Buffer.concat([
                    (0, base_1.tagToBytes)(tag, bigEndian),
                    Buffer.from(vr.name),
                    Buffer.from([0, 0]),
                    (0, base_1.intToBytes)(length, bigEndian),
                ], 12)
            : Buffer.concat([(0, base_1.tagToBytes)(tag, bigEndian), (0, base_1.intToBytes)(length, bigEndian)], 8);
        return new HeaderPart(tag, vr, length, (0, base_1.isFileMetaInformation)(tag), bigEndian, explicitVR, bytes);
    }
    constructor(tag, vr, length, isFmi, bigEndian, explicitVR, bytes) {
        super(bigEndian, bytes);
        this.tag = tag;
        this.vr = vr;
        this.length = length;
        this.isFmi = isFmi;
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
        this.bytes = bytes;
        if (!this.bytes) {
            this.bytes = this.explicitVR
                ? vr.headerLength === 8
                    ? Buffer.concat([(0, base_1.tagToBytes)(tag, bigEndian), Buffer.from(vr.name), (0, base_1.shortToBytes)(length, bigEndian)], 8)
                    : Buffer.concat([
                        (0, base_1.tagToBytes)(tag, bigEndian),
                        Buffer.from(vr.name),
                        Buffer.from([0, 0]),
                        (0, base_1.intToBytes)(length, bigEndian),
                    ], 12)
                : Buffer.concat([(0, base_1.tagToBytes)(tag, bigEndian), (0, base_1.intToBytes)(length, bigEndian)], 8);
        }
    }
    withUpdatedLength(newLength) {
        if (newLength === this.length) {
            return this;
        }
        else {
            let updated = null;
            if (this.bytes.length >= 8 && this.explicitVR && this.vr.headerLength === 8) {
                // explicit vr
                updated = (0, base_1.concat)(this.bytes.slice(0, 6), (0, base_1.shortToBytes)(newLength, this.bigEndian));
            }
            else if (this.bytes.length >= 12 && this.explicitVR && this.vr.headerLength === 12) {
                // explicit vr
                updated = (0, base_1.concat)(this.bytes.slice(0, 8), (0, base_1.intToBytes)(newLength, this.bigEndian));
            }
            else {
                // implicit vr
                updated = (0, base_1.concat)(this.bytes.slice(0, 4), (0, base_1.intToBytes)(newLength, this.bigEndian));
            }
            return new HeaderPart(this.tag, this.vr, newLength, this.isFmi, this.bigEndian, this.explicitVR, updated);
        }
    }
    toString() {
        return ('Header [tag = ' +
            (0, base_1.tagToString)(this.tag) +
            ', vr = ' +
            this.vr.name +
            ', length = ' +
            this.length +
            ', bigEndian = ' +
            this.bigEndian +
            ', explicitVR = ' +
            this.explicitVR +
            ']');
    }
}
exports.HeaderPart = HeaderPart;
class ValueChunk extends DicomPart {
    constructor(bigEndian, bytes, last) {
        super(bigEndian, bytes);
        this.last = last;
    }
    toString() {
        let ascii = (0, base_1.trim)(this.bytes
            .slice(0, 100)
            .toString('ascii')
            .replace(/[^\x20-\x7E]/g, ''));
        if (this.bytes.length > 100) {
            ascii = ascii + '...';
        }
        return 'ValueChunk [length = ' + this.bytes.length + ', last = ' + this.last + ', ascii = ' + ascii + ']';
    }
}
exports.ValueChunk = ValueChunk;
class DeflatedChunk extends DicomPart {
    constructor(bigEndian, bytes) {
        super(bigEndian, bytes);
    }
    toString() {
        return 'DeflatedChunk [length = ' + this.bytes.length + ']';
    }
}
exports.DeflatedChunk = DeflatedChunk;
class ItemPart extends DicomPart {
    constructor(length, bigEndian, bytes) {
        super(bigEndian, bytes);
        this.length = length;
        this.indeterminate = false;
        this.indeterminate = length === base_1.indeterminateLength;
    }
    toString() {
        return 'Item [length = ' + this.length + ']';
    }
}
exports.ItemPart = ItemPart;
class ItemDelimitationPart extends DicomPart {
    constructor(bigEndian, bytes) {
        super(bigEndian, bytes);
    }
    toString() {
        return 'ItemDelimitation';
    }
}
exports.ItemDelimitationPart = ItemDelimitationPart;
class SequencePart extends DicomPart {
    constructor(tag, length, bigEndian, explicitVR, bytes) {
        super(bigEndian, bytes);
        this.tag = tag;
        this.length = length;
        this.explicitVR = explicitVR;
        this.indeterminate = false;
        this.indeterminate = length === base_1.indeterminateLength;
    }
    toString() {
        return 'Sequence [tag = ' + (0, base_1.tagToString)(this.tag) + ', length = ' + this.length + ']';
    }
}
exports.SequencePart = SequencePart;
class SequenceDelimitationPart extends DicomPart {
    constructor(bigEndian, bytes) {
        super(bigEndian, bytes);
    }
    toString() {
        return 'SequenceDelimitation []';
    }
}
exports.SequenceDelimitationPart = SequenceDelimitationPart;
class FragmentsPart extends DicomPart {
    constructor(tag, length, vr, bigEndian, explicitVR, bytes) {
        super(bigEndian, bytes);
        this.tag = tag;
        this.length = length;
        this.vr = vr;
        this.explicitVR = explicitVR;
    }
    toString() {
        return ('Fragments [tag = ' + (0, base_1.tagToString)(this.tag) + ', vr = ' + this.vr.name + ', length = ' + this.length + ']');
    }
}
exports.FragmentsPart = FragmentsPart;
class UnknownPart extends DicomPart {
    constructor(bigEndian, bytes) {
        super(bigEndian, bytes);
    }
    toString() {
        return 'Unknown []';
    }
}
exports.UnknownPart = UnknownPart;
class ElementsPart extends MetaPart {
    constructor(label, elements) {
        super();
        this.label = label;
        this.elements = elements;
    }
}
exports.ElementsPart = ElementsPart;
//# sourceMappingURL=dicom-parts.js.map