"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ElementsBuilder = void 0;
const base_1 = require("./base");
const character_sets_1 = require("./character-sets");
const dicom_elements_1 = require("./dicom-elements");
const tag_1 = require("./tag");
const vr_1 = require("./vr");
const elements_1 = require("./elements");
class DatasetBuilder {
    constructor(characterSets, zoneOffset) {
        this.characterSets = characterSets;
        this.zoneOffset = zoneOffset;
        this.data = new Array(64);
        this.pos = 0;
    }
    addElementSet(elementSet) {
        if (elementSet instanceof dicom_elements_1.ValueElement && elementSet.tag === tag_1.Tag.SpecificCharacterSet) {
            this.characterSets = character_sets_1.CharacterSets.fromBytes(elementSet.value.bytes);
        }
        else if (elementSet instanceof dicom_elements_1.ValueElement && elementSet.tag === tag_1.Tag.TimezoneOffsetFromUTC) {
            const newOffset = (0, elements_1.parseZoneOffset)(elementSet.value.toSingleString(vr_1.VR.SH, elementSet.bigEndian, this.characterSets));
            this.zoneOffset = newOffset || this.zoneOffset;
        }
        if (this.data.length <= this.pos) {
            this.data.length *= 2;
        }
        this.data[this.pos++] = elementSet;
        return this;
    }
    isEmpty() {
        return this.data.length == 0;
    }
    build() {
        return new elements_1.Elements(this.characterSets, this.zoneOffset, this.data.slice(0, this.pos));
    }
}
class ElementsBuilder {
    constructor() {
        this.builderStack = [new DatasetBuilder(base_1.defaultCharacterSet, base_1.systemZone)];
        this.sequenceStack = [];
        this.lengthStack = [];
    }
    addElement(element) {
        if (element == dicom_elements_1.preambleElement && this.builderStack.length == 1 && this.builderStack[0].isEmpty()) {
            return this;
        }
        if (element instanceof dicom_elements_1.ValueElement) {
            this.subtractLength(element.length + (element.explicitVR ? element.vr.headerLength : 8));
            const builder = this.builderStack[0];
            builder.addElementSet(element);
            return this.maybeDelimit();
        }
        if (element instanceof dicom_elements_1.FragmentsElement) {
            this.subtractLength(element.explicitVR ? element.vr.headerLength : 8);
            this.updateFragments(new dicom_elements_1.Fragments(element.tag, element.vr, undefined, [], element.bigEndian, element.explicitVR));
            return this.maybeDelimit();
        }
        if (element instanceof dicom_elements_1.FragmentElement) {
            this.subtractLength(8 + element.length);
            if (this.fragments !== undefined) {
                const updatedFragments = this.fragments.addFragment(new dicom_elements_1.Fragment(element.length, element.value, element.bigEndian));
                this.updateFragments(updatedFragments);
            }
            return this.maybeDelimit();
        }
        if (element instanceof dicom_elements_1.SequenceDelimitationElement && this.hasFragments()) {
            this.subtractLength(8);
            const builder = this.builderStack[0];
            builder.addElementSet(this.fragments);
            this.updateFragments(undefined);
            return this.maybeDelimit();
        }
        if (element instanceof dicom_elements_1.SequenceElement) {
            this.subtractLength(element.explicitVR ? 12 : 8);
            if (!element.indeterminate) {
                this.pushLength(element, element.length);
            }
            this.pushSequence(new dicom_elements_1.Sequence(element.tag, element.indeterminate ? element.length : 0, [], element.bigEndian, element.explicitVR));
            return this.maybeDelimit();
        }
        if (element instanceof dicom_elements_1.ItemElement && this.hasSequence()) {
            this.subtractLength(8);
            const builder = this.builderStack[0];
            const sequence = this.sequenceStack[0].addItem(new dicom_elements_1.Item(elements_1.Elements.empty(), element.indeterminate ? element.length : 0, element.bigEndian));
            if (!element.indeterminate) {
                this.pushLength(element, element.length);
            }
            this.pushBuilder(new DatasetBuilder(builder.characterSets, builder.zoneOffset));
            this.updateSequence(sequence);
            return this.maybeDelimit();
        }
        if (element instanceof dicom_elements_1.ItemDelimitationElement && this.hasSequence()) {
            this.subtractLength(8);
            if (!this.itemIsIndeterminate() && this.lengthStack.length > 0) {
                this.lengthStack.shift(); // determinate length item with delimitation - handle gracefully
            }
            this.endItem();
            return this.maybeDelimit();
        }
        if (element instanceof dicom_elements_1.SequenceDelimitationElement && this.hasSequence()) {
            this.subtractLength(8);
            if (!this.sequenceIsIndeterminate() && this.lengthStack.length > 0) {
                this.lengthStack.shift(); // determinate length sequence with delimitation - handle gracefully
            }
            this.endSequence();
            return this.maybeDelimit();
        }
        console.warn(`Unexpected element ${element}`);
        this.subtractLength(element.toBytes().length);
        return this.maybeDelimit();
    }
    noteElement(element) {
        this.subtractLength(element.toBytes().length);
        return this.maybeDelimit();
    }
    currentDepth() {
        return this.sequenceStack.length;
    }
    build() {
        return this.builderStack.length === 0 ? elements_1.Elements.empty() : this.builderStack[0].build();
    }
    updateSequence(sequence) {
        if (this.sequenceStack.length === 0) {
            this.sequenceStack = [sequence];
        }
        else {
            this.sequenceStack[0] = sequence;
        }
    }
    updateFragments(fragments) {
        this.fragments = fragments;
    }
    subtractLength(length) {
        this.lengthStack.forEach((l) => (l.bytesLeft -= length));
    }
    pushBuilder(builder) {
        this.builderStack.unshift(builder);
    }
    pushSequence(sequence) {
        this.sequenceStack.unshift(sequence);
    }
    pushLength(element, length) {
        this.lengthStack.unshift({ element, bytesLeft: length });
    }
    popBuilder() {
        this.builderStack.shift();
    }
    popSequence() {
        this.sequenceStack.shift();
    }
    hasSequence() {
        return this.sequenceStack.length > 0;
    }
    hasFragments() {
        return this.fragments !== undefined;
    }
    sequenceIsIndeterminate() {
        return this.sequenceStack.length > 0 && this.sequenceStack[0].indeterminate;
    }
    itemIsIndeterminate() {
        return (this.sequenceStack.length > 0 &&
            this.sequenceStack[0].items.length > 0 &&
            this.sequenceStack[0].items[this.sequenceStack[0].items.length - 1].indeterminate);
    }
    endItem() {
        const builder = this.builderStack[0];
        const sequence = this.sequenceStack[0];
        const elements = builder.build();
        const items = sequence.items;
        if (items.length > 0) {
            items[items.length - 1] = items[items.length - 1].setElements(elements);
            const updatedSequence = new dicom_elements_1.Sequence(sequence.tag, sequence.length, items, sequence.bigEndian, sequence.explicitVR);
            this.popBuilder();
            this.updateSequence(updatedSequence);
        }
    }
    endSequence() {
        const sequence = this.sequenceStack[0];
        const sequenceLength = sequence.indeterminate
            ? sequence.length
            : sequence.toBytes().length - (sequence.explicitVR ? 12 : 8);
        const updatedSequence = new dicom_elements_1.Sequence(sequence.tag, sequenceLength, sequence.items, sequence.bigEndian, sequence.explicitVR);
        const builder = this.builderStack[0];
        builder.addElementSet(updatedSequence);
        this.popSequence();
    }
    maybeDelimit() {
        const delimits = this.lengthStack.filter((e) => e.bytesLeft <= 0);
        if (delimits.length > 0) {
            this.lengthStack = this.lengthStack.filter((e) => e.bytesLeft > 0);
            delimits.forEach((e) => {
                if (e.element instanceof dicom_elements_1.ItemElement) {
                    this.endItem();
                }
                else {
                    this.endSequence();
                }
            });
        }
        return this;
    }
}
exports.ElementsBuilder = ElementsBuilder;
//# sourceMappingURL=elements-builder.js.map