"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Elements = exports.parseZoneOffset = void 0;
const js_joda_1 = require("js-joda");
const dicom_elements_1 = require("./dicom-elements");
const base_1 = require("./base");
const character_sets_1 = require("./character-sets");
const lookup_1 = require("./lookup");
const tag_1 = require("./tag");
const tag_path_1 = require("./tag-path");
const value_1 = require("./value");
const vr_1 = require("./vr");
function parseZoneOffset(s) {
    if (s.length < 5) {
        return undefined;
    }
    try {
        return js_joda_1.ZoneOffset.ofTotalMinutes(parseInt(s.slice(0, 1) + (parseInt(s.slice(1, 3), 10) * 60 + parseInt(s.slice(4, 6), 10)), 10));
    }
    catch (error) {
        return undefined;
    }
}
exports.parseZoneOffset = parseZoneOffset;
class Elements {
    static empty(characterSets = base_1.defaultCharacterSet, zoneOffset = base_1.systemZone) {
        return new Elements(characterSets, zoneOffset, []);
    }
    constructor(characterSets = base_1.defaultCharacterSet, zoneOffset = base_1.systemZone, data = []) {
        this.characterSets = characterSets;
        this.zoneOffset = zoneOffset;
        this.data = data;
        this.size = data.length;
    }
    elementByTag(tag) {
        return this.data.find((e) => e.tag === tag);
    }
    elementByPath(tagPath) {
        const tp = tagPath.previous();
        if (tp instanceof tag_path_1.TagPathItem) {
            const e = this.nestedByPath(tp);
            return e === undefined ? undefined : e.elementByTag(tagPath.tag());
        }
        if (tp.isEmpty()) {
            return this.elementByTag(tagPath.tag());
        }
        throw Error('Unsupported tag path type');
    }
    valueElementByTag(tag) {
        return this._valueByTag(tag, (f) => f);
    }
    valueElementByPath(tagPath) {
        return this._valueByPath(tagPath, (f) => f);
    }
    valueByTag(tag) {
        const e = this.valueElementByTag(tag);
        return e ? e.value : undefined;
    }
    valueByPath(tagPath) {
        const e = this.valueElementByPath(tagPath);
        return e ? e.value : undefined;
    }
    bytesByTag(tag) {
        const e = this.valueByTag(tag);
        return e ? e.bytes : undefined;
    }
    bytesByPath(tagPath) {
        const e = this.valueByPath(tagPath);
        return e ? e.bytes : undefined;
    }
    stringsByTag(tag) {
        return this._valuesByTag(tag, (v) => v.value.toStrings(v.vr, v.bigEndian, this.characterSets));
    }
    stringsByPath(tagPath) {
        return this._valuesByPath(tagPath, (v) => v.value.toStrings(v.vr, v.bigEndian, this.characterSets));
    }
    stringByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toString(v.vr, v.bigEndian, this.characterSets));
    }
    stringByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toString(v.vr, v.bigEndian, this.characterSets));
    }
    singleStringByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toSingleString(v.vr, v.bigEndian, this.characterSets));
    }
    singleStringByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toSingleString(v.vr, v.bigEndian, this.characterSets));
    }
    numbersByTag(tag) {
        return this._valuesByTag(tag, (v) => v.value.toNumbers(v.vr, v.bigEndian));
    }
    numbersByPath(tagPath) {
        return this._valuesByPath(tagPath, (v) => v.value.toNumbers(v.vr, v.bigEndian));
    }
    numberByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toNumber(v.vr, v.bigEndian));
    }
    numberByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toNumber(v.vr, v.bigEndian));
    }
    datesByTag(tag) {
        return this._valuesByTag(tag, (v) => v.value.toDates(v.vr));
    }
    datesByPath(tagPath) {
        return this._valuesByPath(tagPath, (v) => v.value.toDates(v.vr));
    }
    dateByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toDate(v.vr));
    }
    dateByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toDate(v.vr));
    }
    timesByTag(tag) {
        return this._valuesByTag(tag, (v) => v.value.toTimes(v.vr));
    }
    timesByPath(tagPath) {
        return this._valuesByPath(tagPath, (v) => v.value.toTimes(v.vr));
    }
    timeByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toTime(v.vr));
    }
    timeByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toTime(v.vr));
    }
    dateTimesByTag(tag) {
        return this._valuesByTag(tag, (v) => v.value.toDateTimes(v.vr, this.zoneOffset));
    }
    dateTimesByPath(tagPath) {
        return this._valuesByPath(tagPath, (v) => v.value.toDateTimes(v.vr, this.zoneOffset));
    }
    dateTimeByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toDateTime(v.vr, this.zoneOffset));
    }
    dateTimeByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toDateTime(v.vr, this.zoneOffset));
    }
    personNamesByTag(tag) {
        return this._valuesByTag(tag, (v) => v.value.toPersonNames(v.vr, this.characterSets));
    }
    personNamesByPath(tagPath) {
        return this._valuesByPath(tagPath, (v) => v.value.toPersonNames(v.vr, this.characterSets));
    }
    personNameByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toPersonName(v.vr, this.characterSets));
    }
    personNameByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toPersonName(v.vr, this.characterSets));
    }
    urlByTag(tag) {
        return this._valueByTag(tag, (v) => v.value.toURL(v.vr));
    }
    urlByPath(tagPath) {
        return this._valueByPath(tagPath, (v) => v.value.toURL(v.vr));
    }
    sequenceByTag(tag) {
        const e = this.elementByTag(tag);
        return e instanceof dicom_elements_1.Sequence ? e : undefined;
    }
    sequenceByPath(tagPath) {
        const e = this._traverseTrunk(this, tagPath.previous());
        return e ? e.sequenceByTag(tagPath.tag()) : undefined;
    }
    itemByTag(tag, index) {
        const s = this.sequenceByTag(tag);
        return s ? s.item(index) : undefined;
    }
    nestedByTag(tag, item) {
        const i = this.itemByTag(tag, item);
        return i ? i.elements : undefined;
    }
    nestedByPath(tagPath) {
        const e = this._traverseTrunk(this, tagPath.previous());
        return e ? e.nestedByTag(tagPath.tag(), tagPath.item) : undefined;
    }
    fragmentsByTag(tag) {
        const e = this.elementByTag(tag);
        return e && e instanceof dicom_elements_1.Fragments ? e : undefined;
    }
    setElementSet(elementSet) {
        if (elementSet instanceof dicom_elements_1.ValueElement && elementSet.tag === tag_1.Tag.SpecificCharacterSet) {
            return new Elements(character_sets_1.CharacterSets.fromBytes(elementSet.value.bytes), this.zoneOffset, this._insertOrdered(elementSet));
        }
        if (elementSet instanceof dicom_elements_1.ValueElement && elementSet.tag === tag_1.Tag.TimezoneOffsetFromUTC) {
            const newOffset = parseZoneOffset(elementSet.value.toString(vr_1.VR.SH));
            const zone = newOffset || base_1.systemZone;
            return new Elements(this.characterSets, zone, this._insertOrdered(elementSet));
        }
        return new Elements(this.characterSets, this.zoneOffset, this._insertOrdered(elementSet));
    }
    setElementSets(elementSets) {
        return elementSets.reduce((elements, elementSet) => elements.setElementSet(elementSet), this);
    }
    setSequence(sequence) {
        return this.setElementSet(sequence);
    }
    updateSequence(tag, index, update) {
        const s1 = this.sequenceByTag(tag);
        if (s1) {
            const i1 = s1.item(index);
            if (i1) {
                const e1 = i1.elements;
                const e2 = update(e1);
                const i2 = i1.setElements(e2);
                const s2 = s1.setItem(index, i2);
                return this.setElementSet(s2);
            }
        }
        return undefined;
    }
    updatePath(elems, tagPath, f) {
        if (tagPath.length === 0) {
            return f(elems);
        }
        const tp = tagPath[0];
        if (tp instanceof tag_path_1.TagPathItem) {
            const updated = elems.updateSequence(tp.tag(), tp.item, (e) => this.updatePath(e, tagPath.slice(1), f));
            return updated ? updated : elems;
        }
        throw Error('Unsupported tag path type');
    }
    setNested(tagPath, elements) {
        return this.updatePath(this, tagPath.toList(), () => elements);
    }
    setNestedElementSet(tagPath, elementSet) {
        return this.updatePath(this, tagPath.toList(), (elements) => elements.setElementSet(elementSet));
    }
    setNestedSequence(tagPath, sequence) {
        return this.setNestedElementSet(tagPath, sequence);
    }
    addItem(tagPath, elements) {
        const sequence = this.sequenceByPath(tagPath);
        if (sequence) {
            const bigEndian = sequence.bigEndian;
            const indeterminate = sequence.indeterminate;
            const item = indeterminate
                ? new dicom_elements_1.Item(elements, base_1.indeterminateLength, bigEndian)
                : new dicom_elements_1.Item(elements, elements.toBytes(false).length, bigEndian);
            const updatedSequence = sequence.addItem(item);
            if (tagPath.previous() === tag_path_1.emptyTagPath) {
                return this.setSequence(updatedSequence);
            }
            else if (tagPath.previous() instanceof tag_path_1.TagPathItem) {
                return this.setNestedSequence(tagPath.previous(), updatedSequence);
            }
            throw Error('Unsupported tag path type');
        }
        return this;
    }
    setCharacterSets(characterSets) {
        return new Elements(characterSets, this.zoneOffset, this.data);
    }
    setZoneOffset(zoneOffset) {
        return new Elements(this.characterSets, zoneOffset, this.data);
    }
    setValue(tag, vr, value, bigEndian = false, explicitVR = true) {
        return this.setElementSet(new dicom_elements_1.ValueElement(tag, vr, value, bigEndian, explicitVR));
    }
    setBytes(tag, vr, value, bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromBuffer(vr, value), bigEndian, explicitVR);
    }
    setStrings(tag, values, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromStrings(vr, values, bigEndian), bigEndian, explicitVR);
    }
    setString(tag, value, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromString(vr, value, bigEndian), bigEndian, explicitVR);
    }
    setNumbers(tag, values, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromNumbers(vr, values, bigEndian), bigEndian, explicitVR);
    }
    setNumber(tag, value, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromNumber(vr, value, bigEndian), bigEndian, explicitVR);
    }
    setDates(tag, values, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromDates(vr, values), bigEndian, explicitVR);
    }
    setDate(tag, value, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromDate(vr, value), bigEndian, explicitVR);
    }
    setTimes(tag, values, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromTimes(vr, values), bigEndian, explicitVR);
    }
    setTime(tag, value, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromTime(vr, value), bigEndian, explicitVR);
    }
    setDateTimes(tag, values, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromDateTimes(vr, values), bigEndian, explicitVR);
    }
    setDateTime(tag, value, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromDateTime(vr, value), bigEndian, explicitVR);
    }
    setPersonNames(tag, values, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromPersonNames(vr, values), bigEndian, explicitVR);
    }
    setPersonName(tag, value, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromPersonName(vr, value), bigEndian, explicitVR);
    }
    setURL(tag, value, vr = lookup_1.Lookup.vrOf(tag), bigEndian = false, explicitVR = true) {
        return this.setValue(tag, vr, value_1.Value.fromURL(vr, value), bigEndian, explicitVR);
    }
    setNestedValue(tagPath, vr, value, bigEndian = false, explicitVR = true) {
        if (tagPath.isRoot()) {
            return this.setValue(tagPath.tag(), vr, value, bigEndian, explicitVR);
        }
        else {
            return this.setNested(tagPath.previous(), (this.nestedByPath(tagPath.previous()) || Elements.empty()).setValue(tagPath.tag(), vr, value, bigEndian, explicitVR));
        }
    }
    setNestedBytes(tagPath, vr, value, bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromBuffer(vr, value), bigEndian, explicitVR);
    }
    setNestedStrings(tagPath, values, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromStrings(vr, values, bigEndian), bigEndian, explicitVR);
    }
    setNestedString(tagPath, value, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromString(vr, value, bigEndian), bigEndian, explicitVR);
    }
    setNestedNumbers(tagPath, values, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromNumbers(vr, values, bigEndian), bigEndian, explicitVR);
    }
    setNestedNumber(tagPath, value, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromNumber(vr, value, bigEndian), bigEndian, explicitVR);
    }
    setNestedDates(tagPath, values, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromDates(vr, values), bigEndian, explicitVR);
    }
    setNestedDate(tagPath, value, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromDate(vr, value), bigEndian, explicitVR);
    }
    setNestedTimes(tagPath, values, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromTimes(vr, values), bigEndian, explicitVR);
    }
    setNestedTime(tagPath, value, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromTime(vr, value), bigEndian, explicitVR);
    }
    setNestedDateTimes(tagPath, values, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromDateTimes(vr, values), bigEndian, explicitVR);
    }
    setNestedDateTime(tagPath, value, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromDateTime(vr, value), bigEndian, explicitVR);
    }
    setNestedPersonNames(tagPath, values, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromPersonNames(vr, values), bigEndian, explicitVR);
    }
    setNestedPersonName(tagPath, value, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromPersonName(vr, value), bigEndian, explicitVR);
    }
    setNestedURL(tagPath, value, vr = lookup_1.Lookup.vrOf(tagPath.tag()), bigEndian = false, explicitVR = true) {
        return this.setNestedValue(tagPath, vr, value_1.Value.fromURL(vr, value), bigEndian, explicitVR);
    }
    removeByTag(tag) {
        return this.filter((elementSet) => elementSet.tag !== tag);
    }
    removeByPath(tagPath) {
        if (tagPath === tag_path_1.emptyTagPath) {
            return this;
        }
        if (tagPath instanceof tag_path_1.TagPathItem) {
            if (tagPath.previous() === tag_path_1.emptyTagPath) {
                const s = this.sequenceByTag(tagPath.tag());
                return s ? this.setSequence(s.removeItem(tagPath.item)) : this;
            }
            if (tagPath.previous() instanceof tag_path_1.TagPathItem) {
                const e = this.nestedByPath(tagPath.previous());
                return e
                    ? this.setNested(tagPath.previous(), e.removeByPath(tag_path_1.TagPath.fromItem(tagPath.tag(), tagPath.item)))
                    : this;
            }
            throw Error('Unsupported tag path type');
        }
        if (tagPath instanceof tag_path_1.TagPathTag) {
            if (tagPath.previous() === tag_path_1.emptyTagPath) {
                return this.removeByTag(tagPath.tag());
            }
            if (tagPath.previous() instanceof tag_path_1.TagPathItem) {
                const e = this.nestedByPath(tagPath.previous());
                return e ? this.setNested(tagPath.previous(), e.removeByTag(tagPath.tag())) : this;
            }
            throw Error('Unsupported tag path type');
        }
        throw Error('Unsupported tag path type');
    }
    filter(f) {
        return new Elements(this.characterSets, this.zoneOffset, this.data.filter(f));
    }
    head() {
        return this.data.length > 0 ? this.data[0] : undefined;
    }
    isEmpty() {
        return this.data.length <= 0;
    }
    nonEmpty() {
        return !this.isEmpty();
    }
    contains(tag) {
        return typeof tag === 'number'
            ? this.data.map((e) => e.tag).includes(tag)
            : tag instanceof tag_path_1.TagPathTag
                ? this.elementByPath(tag) !== undefined
                : tag instanceof tag_path_1.TagPathItem
                    ? this.nestedByPath(tag) !== undefined
                    : false;
    }
    sorted() {
        return new Elements(this.characterSets, this.zoneOffset, this.data.slice().sort((e1, e2) => e1.tag - e2.tag));
    }
    toElements(withPreamble = true) {
        const elements = (0, base_1.flatten)(this.data.map((e) => e.toElements()));
        return withPreamble ? (0, base_1.prependToArray)(dicom_elements_1.preambleElement, elements) : elements;
    }
    toParts(withPreamble) {
        return (0, base_1.flatten)(this.toElements(withPreamble).map((e) => e.toParts()));
    }
    toBytes(withPreamble = true) {
        return this.data
            .map((e) => e.toBytes())
            .reduce((p, e) => (0, base_1.concat)(p, e), withPreamble ? dicom_elements_1.preambleElement.toBytes() : base_1.emptyBuffer);
    }
    toStrings(indent) {
        const space = ' ';
        const space1 = (description) => {
            return space.repeat(Math.max(0, 40 - description.length));
        };
        const space2 = (length) => {
            return space.repeat(Math.max(0, 4 - (length + '').length));
        };
        return (0, base_1.flatten)(this.data.map((e) => {
            if (e instanceof dicom_elements_1.ValueElement) {
                const strings = e.value.toStrings(e.vr, e.bigEndian, this.characterSets);
                const s = strings.join(base_1.multiValueDelimiter);
                const vm = strings.length + '';
                return [
                    indent +
                        (0, base_1.tagToString)(e.tag) +
                        space +
                        e.vr.name +
                        space +
                        s +
                        space +
                        space1(s) +
                        ' # ' +
                        space2(e.length) +
                        space +
                        e.length +
                        ', ' +
                        vm +
                        space +
                        lookup_1.Lookup.keywordOf(e.tag) || '',
                ];
            }
            if (e instanceof dicom_elements_1.Sequence) {
                const hDescription = e.length === base_1.indeterminateLength
                    ? 'Sequence with indeterminate length'
                    : 'Sequence with explicit length ' + e.length;
                const heading = indent +
                    (0, base_1.tagToString)(e.tag) +
                    ' SQ ' +
                    hDescription +
                    space +
                    space1(hDescription) +
                    ' # ' +
                    space2((0, base_1.toInt32)(e.length)) +
                    space +
                    (0, base_1.toInt32)(e.length) +
                    ', 1 ' +
                    lookup_1.Lookup.keywordOf(e.tag) || '';
                const items = (0, base_1.flatten)(e.items.map((i) => {
                    const iDescription = i.indeterminate
                        ? 'Item with indeterminate length'
                        : 'Item with explicit length ' + i.length;
                    const itemHeading = indent +
                        '  ' +
                        (0, base_1.tagToString)(tag_1.Tag.Item) +
                        ' na ' +
                        iDescription +
                        space +
                        space1(iDescription) +
                        ' # ' +
                        space2((0, base_1.toInt32)(i.length)) +
                        space +
                        (0, base_1.toInt32)(i.length) +
                        ', 1 Item';
                    const elems = i.elements.toStrings(indent + '    ');
                    const itemDelimitation = indent +
                        '  ' +
                        (0, base_1.tagToString)(tag_1.Tag.ItemDelimitationItem) +
                        ' na ' +
                        space.repeat(41) +
                        ' #     0, 0 ItemDelimitationItem';
                    elems.unshift(itemHeading);
                    elems.push(itemDelimitation);
                    return elems;
                }));
                const delimitation = indent +
                    (0, base_1.tagToString)(tag_1.Tag.SequenceDelimitationItem) +
                    ' na ' +
                    space.repeat(41) +
                    ' #     0, 0 SequenceDelimitationItem';
                items.unshift(heading);
                items.push(delimitation);
                return items;
            }
            if (e instanceof dicom_elements_1.Fragments) {
                const hDescription = 'Fragments with ' + e.size + ' fragment(s)';
                const heading = indent +
                    (0, base_1.tagToString)(e.tag) +
                    space +
                    e.vr.name +
                    space +
                    hDescription +
                    space +
                    space1(hDescription) +
                    ' #    na, 1 ' +
                    lookup_1.Lookup.keywordOf(e.tag) || '';
                let offsets = [];
                if (e.offsets !== undefined) {
                    const len = e.offsets.length;
                    const description = 'Offsets table with ' + len + ' offset(s)';
                    offsets = [
                        indent +
                            space +
                            space +
                            (0, base_1.tagToString)(tag_1.Tag.Item) +
                            ' na ' +
                            description +
                            space +
                            space1(description) +
                            ' # ' +
                            space2(len * 4) +
                            space +
                            len * 4 +
                            ', 1 Item',
                    ];
                }
                const fragments = e.fragments.map((f) => {
                    const description = 'Fragment with length ' + f.length;
                    return (indent +
                        space +
                        space +
                        (0, base_1.tagToString)(tag_1.Tag.Item) +
                        ' na ' +
                        description +
                        space +
                        space1(description) +
                        ' # ' +
                        space2(f.length) +
                        space +
                        f.length +
                        ', 1 Item');
                });
                const delimitation = indent +
                    (0, base_1.tagToString)(tag_1.Tag.SequenceDelimitationItem) +
                    ' na ' +
                    space.repeat(43) +
                    ' #     0, 0 SequenceDelimitationItem';
                offsets.unshift(heading);
                for (const fragment of fragments) {
                    offsets.push(fragment);
                }
                offsets.push(delimitation);
                return offsets;
            }
            return [];
        }));
    }
    toString() {
        return this.toStrings('').join('\r\n');
    }
    _valueByTag(tag, f) {
        const e = this.elementByTag(tag);
        return e && e instanceof dicom_elements_1.ValueElement ? f(e) : undefined;
    }
    _valueByPath(tagPath, f) {
        const e = this.elementByPath(tagPath);
        return e && e instanceof dicom_elements_1.ValueElement ? f(e) : undefined;
    }
    _valuesByTag(tag, f) {
        const e = this.elementByTag(tag);
        return e && e instanceof dicom_elements_1.ValueElement ? f(e) : [];
    }
    _valuesByPath(tagPath, f) {
        const e = this.elementByPath(tagPath);
        return e && e instanceof dicom_elements_1.ValueElement ? f(e) : [];
    }
    _traverseTrunk(elems, trunk) {
        if (trunk.isEmpty()) {
            return elems;
        }
        else {
            if (trunk instanceof tag_path_1.TagPathItem) {
                const e = this._traverseTrunk(elems, trunk.previous());
                return e ? e.nestedByTag(trunk.tag(), trunk.item) : undefined;
            }
            throw Error('Unsupported tag path type');
        }
    }
    _insertOrdered(element) {
        if (this.isEmpty()) {
            return [element];
        }
        else {
            const b = [];
            let isBelow = true;
            this.data.forEach((e) => {
                if (isBelow && e.tag > element.tag) {
                    b.push(element);
                    isBelow = false;
                }
                if (e.tag === element.tag) {
                    b.push(element);
                    isBelow = false;
                }
                else {
                    b.push(e);
                }
            });
            if (isBelow) {
                b.push(element);
            }
            return b;
        }
    }
}
exports.Elements = Elements;
//# sourceMappingURL=elements.js.map