"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.modifyFlow = exports.TagModificationsPart = exports.TagInsertion = exports.TagModification = void 0;
const base_1 = require("./base");
const dicom_flow_1 = require("./dicom-flow");
const lookup_1 = require("./lookup");
const dicom_parts_1 = require("./dicom-parts");
const tag_path_1 = require("./tag-path");
const vr_1 = require("./vr");
class TagModification {
    static equals(tagPath, modification) {
        return new TagModification(tagPath.isEqualTo.bind(tagPath), modification);
    }
    static endsWith(tagPath, modification) {
        return new TagModification((tp) => tp.endsWith(tagPath), modification);
    }
    constructor(matches, modification) {
        this.matches = matches;
        this.modification = modification;
    }
}
exports.TagModification = TagModification;
class TagInsertion {
    constructor(tagPath, insertion) {
        this.tagPath = tagPath;
        this.insertion = insertion;
        this.tagPath = tagPath;
        this.insertion = insertion;
    }
}
exports.TagInsertion = TagInsertion;
class TagModificationsPart extends dicom_parts_1.MetaPart {
    constructor(modifications = [], insertions = [], replace = false) {
        super();
        this.modifications = modifications;
        this.insertions = insertions;
        this.replace = replace;
    }
}
exports.TagModificationsPart = TagModificationsPart;
function modifyFlow(modifications = [], insertions = [], logGroupLengthWarnings = true) {
    const mods = modifications === undefined ? [] : modifications;
    const irts = insertions === undefined ? [] : insertions;
    const wrns = logGroupLengthWarnings === undefined ? true : logGroupLengthWarnings;
    const organizeInsertions = (inserts) => {
        const distinct = inserts.filter((a, pos, arr) => {
            return arr.findIndex((b) => b.tagPath.isEqualTo(a.tagPath)) === pos;
        }); // distinct by tag path
        return distinct.sort((a, b) => (a.tagPath.isBelow(b.tagPath) ? -1 : 1)); // ordered by tag path
    };
    return (0, dicom_flow_1.createFlow)(new (class extends (0, dicom_flow_1.TagPathTracking)((0, dicom_flow_1.GuaranteedValueEvent)((0, dicom_flow_1.GuaranteedDelimitationEvents)((0, dicom_flow_1.GroupLengthWarnings)((0, dicom_flow_1.InFragments)((0, dicom_flow_1.EndEvent)(dicom_flow_1.DeferToPartFlow)))))) {
        constructor() {
            super();
            this.currentModifications = mods;
            this.currentInsertions = organizeInsertions(irts.slice());
            this.latestTagPath = tag_path_1.emptyTagPath;
            this.value = base_1.emptyBuffer;
            this.bigEndian = false;
            this.explicitVR = true;
            this.setSilent(!wrns);
        }
        onPart(part) {
            if (part instanceof TagModificationsPart) {
                if (part.replace) {
                    this.currentModifications = part.modifications;
                    this.currentInsertions = organizeInsertions(part.insertions.slice());
                }
                else {
                    this.currentModifications = (0, base_1.concatArrays)(this.currentModifications, part.modifications);
                    this.currentInsertions = organizeInsertions((0, base_1.concatArrays)(this.currentInsertions, part.insertions));
                }
                return [];
            }
            if (part instanceof dicom_parts_1.HeaderPart) {
                this.updateSyntax(part);
                const insertParts = this.findInsertParts();
                const modifyPart = this.findModifyPart(part);
                this.latestTagPath = this.tagPath;
                return (0, base_1.concatArrays)(insertParts, modifyPart);
            }
            if (part instanceof dicom_parts_1.SequencePart) {
                const insertParts = this.findInsertParts();
                this.latestTagPath = this.tagPath;
                return (0, base_1.appendToArray)(part, insertParts);
            }
            if (part instanceof dicom_parts_1.ValueChunk) {
                if (this.currentModification !== undefined && this.currentHeader !== undefined) {
                    this.value = (0, base_1.concat)(this.value, part.bytes);
                    if (part.last) {
                        const newValue = (0, base_1.padToEvenLength)(this.currentModification.modification(this.value), this.currentHeader.vr);
                        const newHeader = this.currentHeader.withUpdatedLength(newValue.length);
                        this.currentModification = undefined;
                        this.currentHeader = undefined;
                        return (0, base_1.prependToArray)(newHeader, this.valueOrNot(newValue));
                    }
                    else {
                        return [];
                    }
                }
                else {
                    return [part];
                }
            }
            this.latestTagPath = this.tagPath;
            return [part];
        }
        onEnd() {
            if (this.latestTagPath.isEmpty()) {
                return [];
            }
            else {
                return (0, base_1.flatten)(this.currentInsertions
                    .filter((i) => i.tagPath.isRoot())
                    .filter((m) => this.latestTagPath.isBelow(m.tagPath))
                    .map((m) => this.headerAndValueParts(m.tagPath, (0, base_1.padToEvenLength)(m.insertion(undefined), m.tagPath.tag()))));
            }
        }
        updateSyntax(header) {
            this.bigEndian = header.bigEndian;
            this.explicitVR = header.explicitVR;
        }
        valueOrNot(bytes) {
            return bytes.length > 0 ? [new dicom_parts_1.ValueChunk(this.bigEndian, bytes, true)] : [];
        }
        headerAndValueParts(tagPath, valueBytes) {
            const vr = lookup_1.Lookup.vrOf(tagPath.tag());
            if (vr === vr_1.VR.UN) {
                throw Error('Tag is not present in dictionary, cannot determine value representation');
            }
            if (vr === vr_1.VR.SQ) {
                throw Error('Cannot insert sequences');
            }
            const header = dicom_parts_1.HeaderPart.create(tagPath.tag(), vr, valueBytes.length, this.bigEndian, this.explicitVR);
            return (0, base_1.prependToArray)(header, this.valueOrNot(valueBytes));
        }
        isBetween(lowerTag, tagToTest, upperTag) {
            return lowerTag.isBelow(tagToTest) && tagToTest.isBelow(upperTag);
        }
        isInDataset(tagToTest, tagPath) {
            return tagToTest.previous().isEqualTo(tagPath.previous());
        }
        findInsertParts() {
            return (0, base_1.flatten)(this.currentInsertions
                .filter((i) => this.isBetween(this.latestTagPath, i.tagPath, this.tagPath))
                .filter((i) => this.isInDataset(i.tagPath, this.tagPath))
                .map((i) => this.headerAndValueParts(i.tagPath, (0, base_1.padToEvenLength)(i.insertion(undefined), i.tagPath.tag()))));
        }
        findModifyPart(header) {
            const mod = this.currentModifications.find((m) => m.matches(this.tagPath));
            if (mod !== undefined) {
                this.currentHeader = header;
                this.currentModification = mod;
                this.value = base_1.emptyBuffer;
                return [];
            }
            else {
                const ins = this.currentInsertions.find((i) => i.tagPath.isEqualTo(this.tagPath));
                if (ins !== undefined) {
                    this.currentHeader = header;
                    this.currentModification = new TagModification((tp) => tp.isEqualTo(ins.tagPath), (v) => ins.insertion(v));
                    this.value = base_1.emptyBuffer;
                    return [];
                }
                else {
                    return [header];
                }
            }
        }
    })());
}
exports.modifyFlow = modifyFlow;
//# sourceMappingURL=modify-flow.js.map