"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseFlow = void 0;
const zlib_1 = __importDefault(require("zlib"));
const base_1 = require("./base");
const byte_parser_1 = require("./byte-parser");
const detour_1 = require("./detour");
const lookup_1 = require("./lookup");
const parsing_1 = require("./parsing");
const dicom_parts_1 = require("./dicom-parts");
const tag_1 = require("./tag");
const uid_1 = require("./uid");
const vr_1 = require("./vr");
class DicomParseStep extends byte_parser_1.ParseStep {
    constructor(state, flow) {
        super();
        this.state = state;
        this.flow = flow;
    }
}
class DatasetHeaderState {
    constructor(maySwitchTs, bigEndian, explicitVR) {
        this.maySwitchTs = maySwitchTs;
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
    }
}
class FmiHeaderState {
    constructor(tsuid, bigEndian, explicitVR, hasFmi, pos, fmiEndPos) {
        this.tsuid = tsuid;
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
        this.hasFmi = hasFmi;
        this.pos = pos;
        this.fmiEndPos = fmiEndPos;
    }
}
class ValueState {
    constructor(bigEndian, bytesLeft, nextStep) {
        this.bigEndian = bigEndian;
        this.bytesLeft = bytesLeft;
        this.nextStep = nextStep;
    }
}
class FragmentsState {
    constructor(bigEndian, explicitVR) {
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
    }
}
class AtBeginning extends DicomParseStep {
    constructor(flow) {
        super(null, flow);
    }
    parse(reader) {
        let maybePreamble;
        if (reader.remainingSize() < parsing_1.dicomPreambleLength + 8) {
            if (reader
                .remainingData()
                .slice(0, 128)
                .every((b) => b === 0)) {
                reader.ensure(parsing_1.dicomPreambleLength + 8);
            }
        }
        else if ((0, parsing_1.isPreamble)(reader.remainingData())) {
            maybePreamble = new dicom_parts_1.PreamblePart(reader.take(parsing_1.dicomPreambleLength));
        }
        reader.ensure(8);
        const info = (0, parsing_1.tryReadHeader)(reader.remainingData());
        if (info) {
            let nextState;
            if (info.hasFmi) {
                if (!info.explicitVR) {
                    console.warn('File meta information uses implicit VR encoding');
                }
                if (info.bigEndian) {
                    console.warn('File meta information uses big-endian encoding');
                }
                nextState = new InFmiHeader(new FmiHeaderState(undefined, info.bigEndian, info.explicitVR, info.hasFmi, 0, undefined), this.flow);
            }
            else {
                nextState = new InDatasetHeader(new DatasetHeaderState(false, info.bigEndian, info.explicitVR), this.flow);
            }
            return new byte_parser_1.ParseResult(maybePreamble, nextState);
        }
        else {
            throw new Error('Not a DICOM stream');
        }
    }
    onTruncation(reader) {
        if (reader.remainingSize() === parsing_1.dicomPreambleLength && (0, parsing_1.isPreamble)(reader.remainingData())) {
            this.flow.push(new dicom_parts_1.PreamblePart(reader.take(parsing_1.dicomPreambleLength)));
        }
        else {
            super.onTruncation(reader);
        }
    }
}
class InFmiHeader extends DicomParseStep {
    constructor(state, flow) {
        super(state, flow);
        this.transferSyntaxLengthLimit = 1024;
    }
    toDatasetStep(reader, valueLength) {
        let tsuid = this.state.tsuid;
        if (!tsuid) {
            console.warn('Missing Transfer Syntax (0002,0010) - assume Explicit VR Little Endian');
            tsuid = uid_1.UID.ExplicitVRLittleEndian;
        }
        const bigEndian = tsuid === uid_1.UID.ExplicitVRBigEndianRetired;
        const explicitVR = tsuid !== uid_1.UID.ImplicitVRLittleEndian;
        if ((0, base_1.isDeflated)(tsuid)) {
            if (this.flow.inflate) {
                reader.ensure(valueLength + 2);
                let inflater = zlib_1.default.createInflateRaw();
                const firstTwoBytes = reader.remainingData().slice(valueLength, valueLength + 2);
                const hasZLIBHeader = (0, base_1.bytesToUShortBE)(firstTwoBytes) === 0x789c;
                if (hasZLIBHeader) {
                    console.warn('Deflated DICOM Stream with ZLIB Header');
                    inflater = zlib_1.default.createInflate();
                }
                const valueBytes = reader.take(valueLength);
                const remainingBytes = reader.remainingData();
                reader.setInput(valueBytes);
                this.flow.setDetourFlow(inflater);
                this.flow.setDetour(true, remainingBytes);
            }
            else {
                return new InDeflatedData(this.state, this.flow);
            }
        }
        return new InDatasetHeader(new DatasetHeaderState(true, bigEndian, explicitVR), this.flow);
    }
    parse(reader) {
        const header = (0, parsing_1.readHeader)(reader, this.state);
        (0, parsing_1.warnIfOdd)(header.tag, header.vr, header.valueLength);
        if ((0, base_1.groupNumber)(header.tag) !== 2) {
            console.warn('Missing or wrong File Meta Information Group Length (0002,0000)');
            return new byte_parser_1.ParseResult(undefined, this.toDatasetStep(reader, header.valueLength));
        }
        const updatedVr = header.vr === vr_1.VR.UN ? lookup_1.Lookup.vrOf(header.tag) : header.vr;
        const bytes = reader.take(header.headerLength);
        this.state.pos += header.headerLength + header.valueLength;
        if (header.tag === tag_1.Tag.FileMetaInformationGroupLength) {
            reader.ensure(4);
            const valueBytes = reader.remainingData().slice(0, 4);
            this.state.fmiEndPos = this.state.pos + (0, base_1.bytesToInt)(valueBytes, this.state.bigEndian);
        }
        else if (header.tag === tag_1.Tag.TransferSyntaxUID) {
            if (header.valueLength < this.transferSyntaxLengthLimit) {
                reader.ensure(header.valueLength);
                const valueBytes = reader.remainingData().slice(0, header.valueLength);
                this.state.tsuid = (0, base_1.trim)(valueBytes.toString());
            }
            else {
                console.warn('Transfer syntax data is very large, skipping');
            }
        }
        const part = new dicom_parts_1.HeaderPart(header.tag, updatedVr, header.valueLength, true, this.state.bigEndian, this.state.explicitVR, bytes);
        let nextStep = new InFmiHeader(this.state, this.flow);
        if (this.state.fmiEndPos && this.state.fmiEndPos <= this.state.pos) {
            if (reader.remainingSize() >= header.valueLength + 2 &&
                !(this.state.tsuid && this.state.tsuid == uid_1.UID.DeflatedExplicitVRLittleEndian) &&
                (0, base_1.bytesToShort)(reader.remainingData().slice(header.valueLength, header.valueLength + 2), this.state.bigEndian) == 2) {
                console.warn('Wrong File Meta Information Group Length (0002,0000)');
            }
            else {
                if (this.state.fmiEndPos != this.state.pos) {
                    console.warn('Wrong File Meta Information Group Length (0002,0000)');
                }
                nextStep = this.toDatasetStep(reader, header.valueLength);
            }
        }
        return new byte_parser_1.ParseResult(part, new InValue(new ValueState(this.state.bigEndian, header.valueLength, nextStep), this.flow));
    }
}
class InDatasetHeader extends DicomParseStep {
    constructor(state, flow) {
        super(state, flow);
    }
    maybeSwitchTs(reader, state) {
        reader.ensure(8);
        const data = reader.remainingData().slice(0, 8);
        const tag = (0, base_1.bytesToTag)(data, state.bigEndian);
        let explicitVR = undefined;
        try {
            explicitVR = vr_1.VR.valueOf((0, base_1.bytesToVR)(data.slice(4)));
        }
        catch (e) { }
        if ((0, parsing_1.isSpecial)(tag)) {
            return new DatasetHeaderState(false, state.bigEndian, state.explicitVR);
        }
        if (state.explicitVR && !explicitVR) {
            console.log('Implicit VR attributes detected in explicit VR dataset');
            return new DatasetHeaderState(false, state.bigEndian, false);
        }
        if (!state.explicitVR && explicitVR) {
            return new DatasetHeaderState(false, state.bigEndian, true);
        }
        return new DatasetHeaderState(false, state.bigEndian, state.explicitVR);
    }
    readDatasetHeader(reader, state) {
        const header = (0, parsing_1.readHeader)(reader, state);
        (0, parsing_1.warnIfOdd)(header.tag, header.vr, header.valueLength);
        if (header.vr) {
            const bytes = reader.take(header.headerLength);
            if (header.vr === vr_1.VR.SQ || (header.vr === vr_1.VR.UN && header.valueLength === base_1.indeterminateLength)) {
                return new dicom_parts_1.SequencePart(header.tag, header.valueLength, state.bigEndian, state.explicitVR, bytes);
            }
            if (header.valueLength === base_1.indeterminateLength) {
                return new dicom_parts_1.FragmentsPart(header.tag, header.valueLength, header.vr, state.bigEndian, state.explicitVR, bytes);
            }
            return new dicom_parts_1.HeaderPart(header.tag, header.vr, header.valueLength, false, state.bigEndian, state.explicitVR, bytes);
        }
        switch (header.tag) {
            case 0xfffee000:
                return new dicom_parts_1.ItemPart(header.valueLength, state.bigEndian, reader.take(8));
            case 0xfffee00d:
                return new dicom_parts_1.ItemDelimitationPart(state.bigEndian, reader.take(8));
            case 0xfffee0dd:
                return new dicom_parts_1.SequenceDelimitationPart(state.bigEndian, reader.take(8));
        }
        return new dicom_parts_1.UnknownPart(state.bigEndian, reader.take(header.headerLength));
    }
    parse(reader) {
        const state = this.state.maySwitchTs ? this.maybeSwitchTs(reader, this.state) : this.state;
        const part = this.readDatasetHeader(reader, state);
        let nextState = byte_parser_1.finishedParser;
        if (part) {
            if (part instanceof dicom_parts_1.HeaderPart) {
                if (part.length > 0) {
                    nextState = new InValue(new ValueState(part.bigEndian, part.length, new InDatasetHeader(state, this.flow)), this.flow);
                }
                else {
                    nextState = new InDatasetHeader(state, this.flow);
                }
            }
            else if (part instanceof dicom_parts_1.FragmentsPart) {
                nextState = new InFragments(new FragmentsState(part.bigEndian, state.explicitVR), this.flow);
            }
            else if (part instanceof dicom_parts_1.SequencePart) {
                nextState = new InDatasetHeader(new DatasetHeaderState(false, state.bigEndian, state.explicitVR), this.flow);
            }
            else if (part instanceof dicom_parts_1.ItemPart) {
                nextState = new InDatasetHeader(new DatasetHeaderState(true, state.bigEndian, state.explicitVR), this.flow);
            }
            else if (part instanceof dicom_parts_1.ItemDelimitationPart) {
                nextState = new InDatasetHeader(new DatasetHeaderState(false, state.bigEndian, state.explicitVR), this.flow);
            }
            else if (part instanceof dicom_parts_1.SequenceDelimitationPart) {
                nextState = new InDatasetHeader(new DatasetHeaderState(true, state.bigEndian, state.explicitVR), this.flow);
            }
            else {
                nextState = new InDatasetHeader(new DatasetHeaderState(false, state.bigEndian, state.explicitVR), this.flow);
            }
        }
        return new byte_parser_1.ParseResult(part, nextState);
    }
}
class InValue extends DicomParseStep {
    constructor(state, flow) {
        super(state, flow);
    }
    parse(reader) {
        return this.state.bytesLeft <= this.flow.chunkSize
            ? new byte_parser_1.ParseResult(new dicom_parts_1.ValueChunk(this.state.bigEndian, reader.take(this.state.bytesLeft), true), this.state.nextStep)
            : new byte_parser_1.ParseResult(new dicom_parts_1.ValueChunk(this.state.bigEndian, reader.take(this.flow.chunkSize), false), new InValue(new ValueState(this.state.bigEndian, this.state.bytesLeft - this.flow.chunkSize, this.state.nextStep), this.flow));
    }
}
class InFragments extends DicomParseStep {
    constructor(state, flow) {
        super(state, flow);
    }
    parse(reader) {
        const header = (0, parsing_1.readHeader)(reader, this.state);
        if (header.tag === 0xfffee000) {
            // begin fragment
            const nextState = header.valueLength > 0
                ? new InValue(new ValueState(this.state.bigEndian, header.valueLength, new InFragments(new FragmentsState(this.state.bigEndian, this.state.explicitVR), this.flow)), this.flow)
                : new InFragments(new FragmentsState(this.state.bigEndian, this.state.explicitVR), this.flow);
            return new byte_parser_1.ParseResult(new dicom_parts_1.ItemPart(header.valueLength, this.state.bigEndian, reader.take(header.headerLength)), nextState);
        }
        if (header.tag === 0xfffee0dd) {
            // end fragments
            if (header.valueLength !== 0) {
                console.warn('Unexpected fragments delimitation length ' + header.valueLength);
            }
            return new byte_parser_1.ParseResult(new dicom_parts_1.SequenceDelimitationPart(this.state.bigEndian, reader.take(header.headerLength)), new InDatasetHeader(new DatasetHeaderState(false, this.state.bigEndian, this.state.explicitVR), this.flow));
        }
        console.warn('Unexpected element (' + (0, base_1.tagToString)(header.tag) + ') in fragments with length ' + header.valueLength);
        return new byte_parser_1.ParseResult(new dicom_parts_1.UnknownPart(this.state.bigEndian, reader.take(header.headerLength + header.valueLength)), this);
    }
}
class InDeflatedData extends DicomParseStep {
    constructor(state, flow) {
        super(state, flow);
    }
    parse(reader) {
        return new byte_parser_1.ParseResult(new dicom_parts_1.DeflatedChunk(this.state.bigEndian, reader.take(Math.min(this.flow.chunkSize, reader.remainingSize()))), this);
    }
}
class ParseFlow extends detour_1.Detour {
    constructor(chunkSize = 1024 * 1024, inflate = true, bufferBytes = 1024 * 1024) {
        super({ highWaterMark: bufferBytes, readableObjectMode: true }); // FIXME should be writableHighWaterMark
        this.chunkSize = chunkSize;
        this.inflate = inflate;
        this.bufferBytes = bufferBytes;
        this.parser = new byte_parser_1.ByteParser(this);
        this.parser.startWith(new AtBeginning(this));
    }
    /**
     * Overrides process in Detour. Process a chunk of binary data.
     */
    process(chunk) {
        this.parser.parse(chunk);
    }
    /**
     * Overrides cleanup in Detour. If there are unparsed bytes left, try to parse these, or fail.
     */
    cleanup() {
        this.parser.flush();
    }
    /**
     * Called by byte parser when it is emitting the next parsed element
     */
    next(part) {
        this.push(part);
    }
    /**
     * Called by byte parser when completing parser. Here we signal completion to the stream.
     */
    complete() {
        this.push(null);
    }
    /**
     * Called by byte parser on error. Here we signal error to the stream.
     */
    fail(error) {
        process.nextTick(() => this.emit('error', error));
    }
}
function parseFlow(chunkSize, inflate, bufferBytes) {
    return new ParseFlow(chunkSize, inflate, bufferBytes);
}
exports.parseFlow = parseFlow;
//# sourceMappingURL=parse-flow.js.map