"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Parser = void 0;
const zlib_1 = __importDefault(require("zlib"));
const base_1 = require("./base");
const byte_parser_1 = require("./byte-parser");
const dicom_elements_1 = require("./dicom-elements");
const elements_builder_1 = require("./elements-builder");
const lookup_1 = require("./lookup");
const parsing_1 = require("./parsing");
const tag_1 = require("./tag");
const uid_1 = require("./uid");
const value_1 = require("./value");
const vr_1 = require("./vr");
class Inflater {
    inflate(bytes) {
        return bytes;
    }
}
class FmiAttributeState {
    constructor(tsuid, bigEndian, explicitVR, hasFmi, pos, fmiEndPos) {
        this.tsuid = tsuid;
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
        this.hasFmi = hasFmi;
        this.pos = pos;
        this.fmiEndPos = fmiEndPos;
    }
}
class AttributeState {
    constructor(maySwitchTs, bigEndian, explicitVR, inflater) {
        this.maySwitchTs = maySwitchTs;
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
        this.inflater = inflater;
    }
}
class FragmentsState {
    constructor(bigEndian, explicitVR, inflater) {
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
        this.inflater = inflater;
    }
}
class DicomParseStep extends byte_parser_1.ParseStep {
    constructor(state, stop) {
        super();
        this.state = state;
        this.stop = stop;
    }
}
class AtBeginning extends DicomParseStep {
    constructor(stop) {
        super(null, stop);
    }
    parse(reader) {
        if (reader.remainingSize() < parsing_1.dicomPreambleLength + 8) {
            if (reader
                .remainingData()
                .slice(0, 128)
                .every((b) => b === 0)) {
                reader.ensure(parsing_1.dicomPreambleLength + 8);
            }
        }
        else if ((0, parsing_1.isPreamble)(reader.remainingData())) {
            reader.take(parsing_1.dicomPreambleLength);
        }
        reader.ensure(8);
        const info = (0, parsing_1.tryReadHeader)(reader.remainingData());
        if (info) {
            const nextState = info.hasFmi
                ? new InFmiAttribute(new FmiAttributeState(undefined, info.bigEndian, info.explicitVR, info.hasFmi, 0, undefined), this.stop)
                : new InAttribute(new AttributeState(false, info.bigEndian, info.explicitVR, undefined), this.stop);
            return new byte_parser_1.ParseResult(undefined, nextState);
        }
        else {
            throw new Error('Not a DICOM file');
        }
    }
    onTruncation(reader) {
        if (reader.remainingSize() !== parsing_1.dicomPreambleLength || !(0, parsing_1.isPreamble)(reader.remainingData())) {
            super.onTruncation(reader);
        }
    }
}
class InFmiAttribute extends DicomParseStep {
    constructor(state, stop) {
        super(state, stop);
    }
    parse(reader) {
        const header = (0, parsing_1.readHeader)(reader, this.state);
        if (this.stop(header)) {
            return new byte_parser_1.ParseResult(undefined, byte_parser_1.finishedParser);
        }
        if ((0, base_1.groupNumber)(header.tag) !== 2) {
            console.warn('Missing or wrong File Meta Information Group Length (0002,0000)');
            return new byte_parser_1.ParseResult(undefined, this.toDatasetStep(reader));
        }
        const updatedVr = header.vr === vr_1.VR.UN ? lookup_1.Lookup.vrOf(header.tag) : header.vr;
        const bytes = reader.take(header.headerLength + header.valueLength);
        const valueBytes = bytes.slice(header.headerLength);
        this.state.pos += header.headerLength + header.valueLength;
        if (header.tag === tag_1.Tag.FileMetaInformationGroupLength) {
            this.state.fmiEndPos = this.state.pos + (0, base_1.bytesToInt)(valueBytes, this.state.bigEndian);
        }
        else if (header.tag === tag_1.Tag.TransferSyntaxUID) {
            this.state.tsuid = (0, base_1.trim)(valueBytes.toString());
        }
        return new byte_parser_1.ParseResult(new dicom_elements_1.ValueElement(header.tag, updatedVr, new value_1.Value(valueBytes), this.state.bigEndian, this.state.explicitVR), this.state.fmiEndPos && this.state.fmiEndPos <= this.state.pos ? this.toDatasetStep(reader) : this);
    }
    toDatasetStep(reader) {
        let tsuid = this.state.tsuid;
        if (!tsuid) {
            console.warn('Missing Transfer Syntax (0002,0010) - assume Explicit VR Little Endian');
            tsuid = uid_1.UID.ExplicitVRLittleEndian;
        }
        const bigEndian = tsuid === uid_1.UID.ExplicitVRBigEndianRetired;
        const explicitVR = tsuid !== uid_1.UID.ImplicitVRLittleEndian;
        let inflater;
        if ((0, base_1.isDeflated)(tsuid)) {
            reader.ensure(2);
            inflater = new (class extends Inflater {
                inflate(bytes) {
                    return zlib_1.default.inflateRawSync(bytes);
                }
            })();
            const firstTwoBytes = reader.remainingData().slice(0, 2);
            const hasZLIBHeader = (0, base_1.bytesToUShortBE)(firstTwoBytes) === 0x789c;
            if (hasZLIBHeader) {
                console.warn('Deflated DICOM Stream with ZLIB Header');
                inflater = new (class extends Inflater {
                    inflate(bytes) {
                        return zlib_1.default.inflateSync(bytes);
                    }
                })();
            }
            reader.setInput(inflater.inflate(reader.remainingData()));
        }
        return new InAttribute(new AttributeState(true, bigEndian, explicitVR, inflater), this.stop);
    }
}
class InAttribute extends DicomParseStep {
    constructor(state, stop) {
        super(state, stop);
    }
    maybeSwitchTs(reader, state) {
        reader.ensure(8);
        const data = reader.remainingData().slice(0, 8);
        const tag = (0, base_1.bytesToTag)(data, state.bigEndian);
        let explicitVR = undefined;
        try {
            explicitVR = vr_1.VR.valueOf((0, base_1.bytesToVR)(data.slice(4)));
        }
        catch (e) { }
        if ((0, parsing_1.isSpecial)(tag)) {
            return new AttributeState(false, state.bigEndian, state.explicitVR, state.inflater);
        }
        if (state.explicitVR && !explicitVR) {
            console.log('Implicit VR attributes detected in explicit VR dataset');
            return new AttributeState(false, state.bigEndian, false, state.inflater);
        }
        if (!state.explicitVR && explicitVR) {
            return new AttributeState(false, state.bigEndian, true, state.inflater);
        }
        return new AttributeState(false, state.bigEndian, state.explicitVR, state.inflater);
    }
    parse(reader) {
        const state = this.state.maySwitchTs ? this.maybeSwitchTs(reader, this.state) : this.state;
        const header = (0, parsing_1.readHeader)(reader, state);
        reader.take(header.headerLength);
        if (header.vr) {
            if (this.stop(header)) {
                return new byte_parser_1.ParseResult(undefined, byte_parser_1.finishedParser);
            }
            if (header.vr === vr_1.VR.SQ || (header.vr === vr_1.VR.UN && header.valueLength === base_1.indeterminateLength)) {
                return new byte_parser_1.ParseResult(new dicom_elements_1.SequenceElement(header.tag, header.valueLength, state.bigEndian, state.explicitVR), new InAttribute(new AttributeState(false, state.bigEndian, state.explicitVR, state.inflater), this.stop));
            }
            if (header.valueLength === base_1.indeterminateLength) {
                return new byte_parser_1.ParseResult(new dicom_elements_1.FragmentsElement(header.tag, header.vr, state.bigEndian, state.explicitVR), new InFragments(new FragmentsState(state.bigEndian, state.explicitVR, state.inflater), this.stop));
            }
            return new byte_parser_1.ParseResult(new dicom_elements_1.ValueElement(header.tag, header.vr, new value_1.Value(reader.take(header.valueLength)), state.bigEndian, state.explicitVR), this);
        }
        switch (header.tag) {
            case 0xfffee000:
                return new byte_parser_1.ParseResult(new dicom_elements_1.ItemElement(header.valueLength, state.bigEndian), new InAttribute(new AttributeState(true, state.bigEndian, state.explicitVR, state.inflater), this.stop));
            case 0xfffee00d:
                return new byte_parser_1.ParseResult(new dicom_elements_1.ItemDelimitationElement(state.bigEndian), new InAttribute(new AttributeState(false, state.bigEndian, state.explicitVR, state.inflater), this.stop));
            case 0xfffee0dd:
                return new byte_parser_1.ParseResult(new dicom_elements_1.SequenceDelimitationElement(state.bigEndian), new InAttribute(new AttributeState(true, state.bigEndian, state.explicitVR, state.inflater), this.stop));
        }
        return new byte_parser_1.ParseResult(new dicom_elements_1.UnknownElement(state.bigEndian), this);
    }
}
class InFragments extends DicomParseStep {
    constructor(state, stop) {
        super(state, stop);
    }
    parse(reader) {
        const header = (0, parsing_1.readHeader)(reader, this.state);
        reader.take(header.headerLength);
        if (header.tag === 0xfffee000) {
            // begin fragment
            const valueBytes = reader.take(header.valueLength);
            return new byte_parser_1.ParseResult(new dicom_elements_1.FragmentElement(header.valueLength, new value_1.Value(valueBytes), this.state.bigEndian), new InFragments(new FragmentsState(this.state.bigEndian, this.state.explicitVR, this.state.inflater), this.stop));
        }
        if (header.tag === 0xfffee0dd) {
            // end fragments
            if (header.valueLength !== 0) {
                console.warn('Unexpected fragments delimitation length ' + header.valueLength);
            }
            return new byte_parser_1.ParseResult(new dicom_elements_1.SequenceDelimitationElement(this.state.bigEndian), new InAttribute(new AttributeState(false, this.state.bigEndian, this.state.explicitVR, this.state.inflater), this.stop));
        }
        reader.take(header.valueLength);
        console.warn('Unexpected element (' + (0, base_1.tagToString)(header.tag) + ') in fragments with length ' + header.valueLength);
        return new byte_parser_1.ParseResult(new dicom_elements_1.UnknownElement(this.state.bigEndian), this);
    }
}
class Parser {
    constructor(stop) {
        this.stop = stop;
        this.builder = new elements_builder_1.ElementsBuilder();
        this.byteParser = new byte_parser_1.ByteParser(this);
        const shouldStop = stop
            ? (attributeInfo) => stop(attributeInfo, this.builder.currentDepth())
            : () => false;
        this.byteParser.startWith(new AtBeginning(shouldStop));
    }
    /**
     * Parse the input binary data, producing DICOM elements that are added to the internal builder. An elements
     * structure can be fetched based on the builder at any time.
     */
    parse(chunk) {
        const step = this.byteParser.current instanceof DicomParseStep ? this.byteParser.current : undefined;
        if (step && step.state && step.state.inflater) {
            chunk = step.state.inflater.inflate(chunk);
        }
        this.byteParser.parse(chunk);
    }
    /**
     * Called by byte parser when it is emitting the next parsed element.
     */
    next(element) {
        this.builder.addElement(element);
    }
    /**
     * Get the current elements as represented by the builder
     */
    result() {
        return this.builder.build();
    }
    /**
     * Returns true the parser has completed parsing (stop condition was met)
     */
    isComplete() {
        return this.byteParser.isCompleted;
    }
    /**
     * Called by byte parser when completing parser. Nothing to do here.
     */
    complete() {
        // do nothing
    }
    /**
     * Called by byte parser on error. Here we just throw the error.
     */
    fail(error) {
        throw error;
    }
}
exports.Parser = Parser;
//# sourceMappingURL=parser.js.map