"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.warnIfOdd = exports.readHeader = exports.AttributeInfo = exports.readTagVr = exports.isSpecial = exports.TagVr = exports.isPreamble = exports.headerInfo = exports.tryReadHeader = exports.HeaderInfo = exports.isDICM = exports.dicomPreambleLength = void 0;
const base_1 = require("./base");
const lookup_1 = require("./lookup");
const vr_1 = require("./vr");
exports.dicomPreambleLength = 132;
function isDICM(bytes) {
    return bytes[0] === 68 && bytes[1] === 73 && bytes[2] === 67 && bytes[3] === 77;
}
exports.isDICM = isDICM;
class HeaderInfo {
    constructor(bigEndian, explicitVR, hasFmi) {
        this.bigEndian = bigEndian;
        this.explicitVR = explicitVR;
        this.hasFmi = hasFmi;
    }
}
exports.HeaderInfo = HeaderInfo;
function tryReadHeader(data) {
    const info = headerInfo(data, false);
    return info === undefined ? headerInfo(data, true) : info;
}
exports.tryReadHeader = tryReadHeader;
function headerInfo(data, assumeBigEndian) {
    const tag = (0, base_1.bytesToTag)(data, assumeBigEndian);
    const vr = lookup_1.Lookup.vrOf(tag);
    if (vr === vr_1.VR.UN || ((0, base_1.groupNumber)(tag) !== 2 && (0, base_1.groupNumber)(tag) < 8)) {
        return undefined;
    }
    if ((0, base_1.bytesToVR)(data.slice(4, 6)) === vr.code) {
        return {
            bigEndian: assumeBigEndian,
            explicitVR: true,
            hasFmi: (0, base_1.isFileMetaInformation)(tag),
        };
    }
    if ((0, base_1.bytesToUInt)(data.slice(4, 8), assumeBigEndian) >= 0) {
        if (assumeBigEndian) {
            throw Error('Implicit VR Big Endian encoded DICOM Stream');
        }
        else {
            return {
                bigEndian: false,
                explicitVR: false,
                hasFmi: (0, base_1.isFileMetaInformation)(tag),
            };
        }
    }
    return undefined;
}
exports.headerInfo = headerInfo;
function isPreamble(data) {
    return data.length >= exports.dicomPreambleLength && isDICM(data.slice(exports.dicomPreambleLength - 4, exports.dicomPreambleLength));
}
exports.isPreamble = isPreamble;
class TagVr {
    constructor(tag, vr) {
        this.tag = tag;
        this.vr = vr;
    }
}
exports.TagVr = TagVr;
function isSpecial(tag) {
    return tag === 0xfffee000 || tag === 0xfffee00d || tag === 0xfffee0dd;
}
exports.isSpecial = isSpecial;
function readTagVr(data, bigEndian, explicitVr) {
    const tag = (0, base_1.bytesToTag)(data, bigEndian);
    if (isSpecial(tag)) {
        return new TagVr(tag, undefined);
    }
    if (explicitVr) {
        return new TagVr(tag, vr_1.VR.valueOf((0, base_1.bytesToVR)(data.slice(4, 6))));
    }
    return new TagVr(tag, lookup_1.Lookup.vrOf(tag));
}
exports.readTagVr = readTagVr;
class AttributeInfo {
    constructor(tag, vr, headerLength, valueLength) {
        this.tag = tag;
        this.vr = vr;
        this.headerLength = headerLength;
        this.valueLength = valueLength;
    }
}
exports.AttributeInfo = AttributeInfo;
function readHeader(reader, state) {
    reader.ensure(8);
    const tagVrBytes = reader.remainingData().slice(0, 8);
    const tagVr = readTagVr(tagVrBytes, state.bigEndian, state.explicitVR);
    if (tagVr.vr && state.explicitVR) {
        if (tagVr.vr.headerLength === 8) {
            return new AttributeInfo(tagVr.tag, tagVr.vr, 8, (0, base_1.bytesToUShort)(tagVrBytes.slice(6), state.bigEndian));
        }
        reader.ensure(12);
        return new AttributeInfo(tagVr.tag, tagVr.vr, 12, (0, base_1.bytesToUInt)(reader.remainingData().slice(8), state.bigEndian));
    }
    return new AttributeInfo(tagVr.tag, tagVr.vr, 8, (0, base_1.bytesToUInt)(tagVrBytes.slice(4), state.bigEndian));
}
exports.readHeader = readHeader;
function warnIfOdd(tag, vr, valueLength) {
    if (valueLength % 2 > 0 && valueLength != base_1.indeterminateLength && vr != null && vr != vr_1.VR.SQ) {
        console.warn(`Element ${(0, base_1.tagToString)(tag)} has odd length`);
    }
}
exports.warnIfOdd = warnIfOdd;
//# sourceMappingURL=parsing.js.map