"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TagPathItemEnd = exports.TagPathItem = exports.TagPathSequenceEnd = exports.TagPathSequence = exports.TagPathTag = exports.emptyTagPath = exports.TagPathTrunk = exports.TagPath = void 0;
const base_1 = require("./base");
const lookup_1 = require("./lookup");
const tag_path_like_1 = require("./tag-path-like");
class TagPath extends tag_path_like_1.TagPathLike {
    static fromTag(tag) {
        return exports.emptyTagPath.thenTag(tag);
    }
    static fromSequence(tag) {
        return exports.emptyTagPath.thenSequence(tag);
    }
    static fromSequenceEnd(tag) {
        return exports.emptyTagPath.thenSequenceEnd(tag);
    }
    static fromItem(tag, item) {
        return exports.emptyTagPath.thenItem(tag, item);
    }
    static fromItemEnd(tag, item) {
        return exports.emptyTagPath.thenItemEnd(tag, item);
    }
    static parse(s) {
        const indexPart = (s1) => s1.substring(s1.lastIndexOf('[') + 1, s1.length - 1);
        const tagPart = (s1) => s1.substring(0, s1.indexOf('['));
        const parseTag = (s1) => {
            const tag = lookup_1.Lookup.tagOf(s1);
            if (!tag) {
                if (s1.length === 11 && s1[0] === '(' && s1[5] === ',' && s1[10] === ')') {
                    const i = parseInt(s1.substring(1, 5) + s1.substring(6, 10), 16);
                    if (!isNaN(i)) {
                        return i;
                    }
                }
                throw Error(s1 + ' is not a tag or name string');
            }
            return tag;
        };
        const parseIndex = (s1) => {
            const i = parseInt(s1, 10);
            if (isNaN(i)) {
                throw Error(s1 + ' is not a number');
            }
            return i;
        };
        const createTag = (s1) => TagPath.fromTag(parseTag(s1));
        const addTag = (s1, path) => path.thenTag(parseTag(s1));
        const createSeq = (s1) => TagPath.fromItem(parseTag(tagPart(s1)), parseIndex(indexPart(s1)));
        const addSeq = (s1, path) => path.thenItem(parseTag(tagPart(s1)), parseIndex(indexPart(s1)));
        const tags = s.indexOf('.') > 0 ? s.split('.') : [s];
        const seqTags = tags.length > 1 ? tags.slice(0, tags.length - 1) : []; // list of sequence tags, if any
        const lastTag = tags[tags.length - 1]; // tag or sequence
        try {
            const first = seqTags.length > 0 ? seqTags[0] : undefined;
            if (first) {
                const path = seqTags.slice(1, seqTags.length).reduce((p, tag) => addSeq(tag, p), createSeq(first));
                return addTag(lastTag, path);
            }
            return createTag(lastTag);
        }
        catch (error) {
            throw Error('Tag path could not be parsed: ' + error.message);
        }
    }
    constructor(tag, previous) {
        super();
        this.tagVal = tag;
        this.previousVal = previous;
    }
    tag() {
        return this.tagVal;
    }
    previous() {
        return this.previousVal;
    }
    isEmpty() {
        return this === exports.emptyTagPath;
    }
    isBelow(that) {
        const thisList = this.toList();
        const thatList = that.toList();
        for (let i = 0; i < Math.min(thisList.length, thatList.length); i++) {
            const thisPath = thisList[i];
            const thatPath = thatList[i];
            if (thisPath.isEmpty()) {
                return !thatPath.isEmpty();
            }
            if (thatPath.isEmpty()) {
                return false;
            }
            if (thisPath.tag() !== thatPath.tag()) {
                return thisPath.tag() < thatPath.tag();
            }
            if (thisPath instanceof TagPathSequence && 'item' in thatPath) {
                return true;
            }
            if (thisPath instanceof TagPathSequence && thatPath instanceof TagPathSequenceEnd) {
                return true;
            }
            if (thisPath instanceof TagPathSequenceEnd && 'item' in thatPath) {
                return false;
            }
            if (thisPath instanceof TagPathSequenceEnd && thatPath instanceof TagPathSequence) {
                return false;
            }
            if ('item' in thisPath && thatPath instanceof TagPathSequence) {
                return false;
            }
            if ('item' in thisPath && thatPath instanceof TagPathSequenceEnd) {
                return true;
            }
            if ('item' in thisPath &&
                'item' in thatPath &&
                thisPath.item !== thatPath.item) {
                return thisPath.item < thatPath.item;
            }
            if (thisPath instanceof TagPathItem && thatPath instanceof TagPathItemEnd) {
                return true;
            }
            if (thisPath instanceof TagPathItemEnd && thatPath instanceof TagPathItem) {
                return false;
            }
        }
        return thisList.length < thatList.length;
    }
    isEqualTo(that) {
        if (this.isEmpty() && that.isEmpty()) {
            return true;
        }
        if (this instanceof TagPathTag && that instanceof TagPathTag) {
            return this.tag() === that.tag() && this.previous().isEqualTo(that.previous());
        }
        if (this instanceof TagPathSequence && that instanceof TagPathSequence) {
            return this.tag() === that.tag() && this.previous().isEqualTo(that.previous());
        }
        if (this instanceof TagPathSequenceEnd && that instanceof TagPathSequenceEnd) {
            return this.tag() === that.tag() && this.previous().isEqualTo(that.previous());
        }
        if (this instanceof TagPathItem && that instanceof TagPathItem) {
            return this.tag() === that.tag() && this.item === that.item && this.previous().isEqualTo(that.previous());
        }
        if (this instanceof TagPathItemEnd && that instanceof TagPathItemEnd) {
            return this.tag() === that.tag() && this.item === that.item && this.previous().isEqualTo(that.previous());
        }
        return false;
    }
    startsWith(that) {
        const thisDepth = this.depth();
        const thatDepth = that.depth();
        if (thisDepth >= thatDepth) {
            const n = Math.min(thisDepth, thatDepth);
            return this.take(n).isEqualTo(that.take(n));
        }
        else {
            return false;
        }
    }
    endsWith(that) {
        const n = this.depth() - that.depth();
        return n >= 0 ? this.drop(n).isEqualTo(that) : false;
    }
    drop(n) {
        const dropRec = (path, i) => {
            if (i < 0) {
                return exports.emptyTagPath;
            }
            if (i === 0) {
                if (path.isEmpty()) {
                    return exports.emptyTagPath;
                }
                if (path instanceof TagPathItem) {
                    return TagPath.fromItem(path.tag(), path.item);
                }
                if (path instanceof TagPathItemEnd) {
                    return TagPath.fromItemEnd(path.tag(), path.item);
                }
                if (path instanceof TagPathSequence) {
                    return TagPath.fromSequence(path.tag());
                }
                if (path instanceof TagPathSequenceEnd) {
                    return TagPath.fromSequenceEnd(path.tag());
                }
                return TagPath.fromTag(path.tag());
            }
            const p = dropRec(path.previous(), i - 1);
            if (path instanceof TagPathItem) {
                return p.thenItem(path.tag(), path.item);
            }
            if (path instanceof TagPathItemEnd) {
                return p.thenItemEnd(path.tag(), path.item);
            }
            if (path instanceof TagPathSequence) {
                return p.thenSequence(path.tag());
            }
            if (path instanceof TagPathSequenceEnd) {
                return p.thenSequenceEnd(path.tag());
            }
            if (path instanceof TagPathTag) {
                return p.thenTag(path.tag());
            }
            return exports.emptyTagPath;
        };
        return dropRec(this, this.depth() - n - 1);
    }
    toNamedString(lookup) {
        const toTagString = (tag) => {
            if (lookup) {
                const keyword = lookup_1.Lookup.keywordOf(tag);
                if (keyword) {
                    return keyword;
                }
            }
            return (0, base_1.tagToString)(tag);
        };
        const toTagPathString = (path, tail) => {
            const itemIndexSuffix = 'item' in path ? '[' + path.item + ']' : '';
            const head = toTagString(path.tag()) + itemIndexSuffix;
            const part = head + tail;
            return path.isRoot() ? part : toTagPathString(path.previous(), '.' + part);
        };
        return this.isEmpty() ? '<empty path>' : toTagPathString(this, '');
    }
}
exports.TagPath = TagPath;
class TagPathTrunk extends TagPath {
    thenTag(tag) {
        return new TagPathTag(tag, this);
    }
    thenSequence(tag) {
        return new TagPathSequence(tag, this);
    }
    thenSequenceEnd(tag) {
        return new TagPathSequenceEnd(tag, this);
    }
    thenItem(tag, item) {
        return new TagPathItem(tag, item, this);
    }
    thenItemEnd(tag, item) {
        return new TagPathItemEnd(tag, item, this);
    }
}
exports.TagPathTrunk = TagPathTrunk;
class EmptyTagPath extends TagPathTrunk {
    tag() {
        throw Error('Empty tag path');
    }
    previous() {
        return exports.emptyTagPath;
    }
}
exports.emptyTagPath = new EmptyTagPath(-1, null);
class TagPathTag extends TagPath {
    constructor(tag, previous) {
        super(tag, previous);
    }
}
exports.TagPathTag = TagPathTag;
class TagPathSequence extends TagPath {
    constructor(tag, previous) {
        super(tag, previous);
    }
}
exports.TagPathSequence = TagPathSequence;
class TagPathSequenceEnd extends TagPath {
    constructor(tag, previous) {
        super(tag, previous);
    }
}
exports.TagPathSequenceEnd = TagPathSequenceEnd;
class TagPathItem extends TagPathTrunk {
    constructor(tag, item, previous) {
        super(tag, previous);
        this.item = item;
    }
}
exports.TagPathItem = TagPathItem;
class TagPathItemEnd extends TagPathTrunk {
    constructor(tag, item, previous) {
        super(tag, previous);
        this.item = item;
    }
}
exports.TagPathItemEnd = TagPathItemEnd;
//# sourceMappingURL=tag-path.js.map