"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TagTreeItem = exports.TagTreeAnyItem = exports.TagTreeTag = exports.emptyTagTree = exports.TagTreeTrunk = exports.TagTree = void 0;
const base_1 = require("./base");
const lookup_1 = require("./lookup");
const tag_path_1 = require("./tag-path");
const tag_path_like_1 = require("./tag-path-like");
class TagTree extends tag_path_like_1.TagPathLike {
    static fromTag(tag) {
        return exports.emptyTagTree.thenTag(tag);
    }
    static fromAnyItem(tag) {
        return exports.emptyTagTree.thenAnyItem(tag);
    }
    static fromItem(tag, item) {
        return exports.emptyTagTree.thenItem(tag, item);
    }
    static fromPath(tagPath) {
        let root;
        const p = tagPath.head();
        if (p instanceof tag_path_1.TagPathTag) {
            root = TagTree.fromTag(p.tag());
        }
        else if (p instanceof tag_path_1.TagPathItem) {
            root = TagTree.fromItem(p.tag(), p.item);
        }
        else if (p instanceof tag_path_1.TagPathItemEnd) {
            root = TagTree.fromItem(p.tag(), p.item);
        }
        else if (p instanceof tag_path_1.TagPathSequence) {
            root = TagTree.fromAnyItem(p.tag());
        }
        else if (p instanceof tag_path_1.TagPathSequenceEnd) {
            root = TagTree.fromAnyItem(p.tag());
        }
        else {
            root = exports.emptyTagTree;
        }
        return tagPath
            .drop(1)
            .toList()
            .reduce((t, p1) => {
            if (t instanceof TagTreeTrunk && p1 instanceof tag_path_1.TagPathTag) {
                return t.thenTag(p1.tag());
            }
            if (t instanceof TagTreeTrunk && p1 instanceof tag_path_1.TagPathItem) {
                return t.thenItem(p1.tag(), p1.item);
            }
            if (t instanceof TagTreeTrunk && p1 instanceof tag_path_1.TagPathItemEnd) {
                return t.thenItem(p1.tag(), p1.item);
            }
            if (t instanceof TagTreeTrunk && p1 instanceof tag_path_1.TagPathSequence) {
                return t.thenAnyItem(p1.tag());
            }
            if (t instanceof TagTreeTrunk && p1 instanceof tag_path_1.TagPathSequenceEnd) {
                return t.thenAnyItem(p1.tag());
            }
            return t;
        }, root);
    }
    static parse(str) {
        const isSeq = (s) => s[s.length - 1] === ']';
        const indexPart = (s) => s.substring(s.lastIndexOf('[') + 1, s.length - 1);
        const tagPart = (s) => s.substring(0, s.indexOf('['));
        const parseTag = (s) => {
            const tag = lookup_1.Lookup.tagOf(s);
            if (!tag) {
                if (s.length === 11 && s[0] === '(' && s[5] === ',' && s[10] === ')') {
                    const i = parseInt(s.substring(1, 5) + s.substring(6, 10), 16);
                    if (!isNaN(i)) {
                        return i;
                    }
                }
                throw Error(s + ' is not a tag or name string');
            }
            return tag;
        };
        const parseIndex = (s) => {
            if (s === '*') {
                return undefined;
            }
            const i = parseInt(s, 10);
            if (isNaN(i)) {
                throw Error(s + ' is not a number');
            }
            return i;
        };
        const createTag = (s) => TagTree.fromTag(parseTag(s));
        const addTag = (s, path) => path.thenTag(parseTag(s));
        const createSeq = (s) => {
            const tag = parseTag(tagPart(s));
            const index = parseIndex(indexPart(s));
            return index === undefined ? TagTree.fromAnyItem(tag) : TagTree.fromItem(tag, index);
        };
        const addSeq = (s, path) => {
            const tag = parseTag(tagPart(s));
            const index = parseIndex(indexPart(s));
            return index === undefined ? path.thenAnyItem(tag) : path.thenItem(tag, index);
        };
        const tags = str.indexOf('.') > 0 ? str.split('.') : [str];
        const seqTags = tags.length > 1 ? tags.slice(0, tags.length - 1) : []; // list of sequence tags, if any
        const lastTag = tags[tags.length - 1]; // tag or sequence
        try {
            const first = seqTags.length > 0 ? seqTags[0] : undefined;
            if (first) {
                const tree = seqTags
                    .slice(1, seqTags.length)
                    .reduce((tr, tag) => addSeq(tag, tr), createSeq(first));
                if (tree !== undefined) {
                    return isSeq(lastTag) ? addSeq(lastTag, tree) : addTag(lastTag, tree);
                }
                return isSeq(lastTag) ? createSeq(lastTag) : createTag(lastTag);
            }
            return createTag(lastTag);
        }
        catch (error) {
            throw Error('Tag tree could not be parsed: ' + error.message);
        }
    }
    constructor(tag, previous) {
        super();
        this.tagVal = tag;
        this.previousVal = previous;
    }
    tag() {
        return this.tagVal;
    }
    previous() {
        return this.previousVal;
    }
    isEmpty() {
        return this === exports.emptyTagTree;
    }
    isEqualTo(that) {
        if (this.isEmpty() && that.isEmpty()) {
            return true;
        }
        if (this instanceof TagTreeTag && that instanceof TagTreeTag) {
            return this.tag() === that.tag() && this.previous().isEqualTo(that.previous());
        }
        if (this instanceof TagTreeItem && that instanceof TagTreeItem) {
            return this.tag() === that.tag() && this.item === that.item && this.previous().isEqualTo(that.previous());
        }
        if (this instanceof TagTreeAnyItem && that instanceof TagTreeAnyItem) {
            return this.tag() === that.tag() && this.previous().isEqualTo(that.previous());
        }
        return false;
    }
    isPath() {
        if (this.isEmpty()) {
            return true;
        }
        if (this instanceof TagTreeAnyItem) {
            return false;
        }
        return this.previous().isPath();
    }
    hasPath(tagPath) {
        if (this.isEmpty() && tagPath.isEmpty()) {
            return true;
        }
        if (this instanceof TagTreeTag && tagPath instanceof tag_path_1.TagPathTag) {
            return this.tag() === tagPath.tag() && this.previous().hasPath(tagPath.previous());
        }
        if (this instanceof TagTreeItem && 'item' in tagPath) {
            return (this.item === tagPath.item &&
                this.tag() === tagPath.tag() &&
                this.previous().hasPath(tagPath.previous()));
        }
        if (this instanceof TagTreeAnyItem && 'item' in tagPath) {
            return (this.tag() === tagPath.tag() && this.previous().hasPath(tagPath.previous()));
        }
        if (this instanceof TagTreeAnyItem && tagPath instanceof tag_path_1.TagPathSequence) {
            return this.tag() === tagPath.tag() && this.previous().hasPath(tagPath.previous());
        }
        if (this instanceof TagTreeAnyItem && tagPath instanceof tag_path_1.TagPathSequenceEnd) {
            return this.tag() === tagPath.tag() && this.previous().hasPath(tagPath.previous());
        }
        return false;
    }
    hasTrunk(tagPath) {
        if (this.depth() >= tagPath.depth()) {
            const thisList = this.toList();
            const thatList = tagPath.toList();
            for (let i = 0; i < Math.min(thisList.length, thatList.length); i++) {
                const t = thisList[i];
                const p = thatList[i];
                let check = false;
                if (p.isEmpty()) {
                    check = true;
                }
                else if (t instanceof TagTreeItem && 'item' in p) {
                    check = t.tag() === p.tag() && t.item === p.item;
                }
                else if (t instanceof TagTreeItem && p instanceof tag_path_1.TagPathSequence) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeItem && p instanceof tag_path_1.TagPathSequenceEnd) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeAnyItem && p.item !== undefined) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeAnyItem && p instanceof tag_path_1.TagPathSequence) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeAnyItem && p instanceof tag_path_1.TagPathSequenceEnd) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeTag && p instanceof tag_path_1.TagPathTag) {
                    check = t.tag() === p.tag();
                }
                if (!check) {
                    return false;
                }
            }
            return true;
        }
        else {
            return false;
        }
    }
    isTrunkOf(tagPath) {
        if (this.depth() <= tagPath.depth()) {
            const thisList = this.toList();
            const thatList = tagPath.toList();
            for (let i = 0; i < Math.min(thisList.length, thatList.length); i++) {
                const t = thisList[i];
                const p = thatList[i];
                let check = false;
                if (p.isEmpty()) {
                    check = true;
                }
                else if (t instanceof TagTreeItem && 'item' in p) {
                    check = t.tag() === p.tag() && t.item === p.item;
                }
                else if (t instanceof TagTreeAnyItem && 'item' in p) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeAnyItem && p instanceof tag_path_1.TagPathSequence) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeAnyItem && p instanceof tag_path_1.TagPathSequenceEnd) {
                    check = t.tag() === p.tag();
                }
                else if (t instanceof TagTreeTag && p instanceof tag_path_1.TagPathTag) {
                    check = t.tag() === p.tag();
                }
                if (!check) {
                    return false;
                }
            }
            return true;
        }
        else {
            return false;
        }
    }
    hasTwig(tagPath) {
        let check = false;
        if (this.isEmpty() && tagPath.isEmpty()) {
            check = true;
        }
        else if (this instanceof TagTreeAnyItem && 'item' in tagPath) {
            check = this.tag() === tagPath.tag();
        }
        else if (this instanceof TagTreeAnyItem && tagPath instanceof tag_path_1.TagPathSequence) {
            check = this.tag() === tagPath.tag();
        }
        else if (this instanceof TagTreeAnyItem && tagPath instanceof tag_path_1.TagPathSequenceEnd) {
            check = this.tag() === tagPath.tag();
        }
        else if (this instanceof TagTreeItem && 'item' in tagPath) {
            check = this.tag() === tagPath.tag() && this.item === tagPath.item;
        }
        else if (this instanceof TagTreeTag && tagPath instanceof tag_path_1.TagPathTag) {
            check = this.tag() === tagPath.tag();
        }
        if (tagPath.previous().isEmpty()) {
            return check;
        }
        else if (this.previous().isEmpty()) {
            return false;
        }
        return check && this.previous().hasTwig(tagPath.previous());
    }
    drop(n) {
        const dropRec = (tree, i) => {
            if (i < 0) {
                return exports.emptyTagTree;
            }
            if (i === 0) {
                if (tree.isEmpty()) {
                    return exports.emptyTagTree;
                }
                if (tree instanceof TagTreeItem) {
                    return TagTree.fromItem(tree.tag(), tree.item);
                }
                if (tree instanceof TagTreeAnyItem) {
                    return TagTree.fromAnyItem(tree.tag());
                }
                return TagTree.fromTag(tree.tag());
            }
            const t = dropRec(tree.previous(), i - 1);
            if (tree instanceof TagTreeItem) {
                return t.thenItem(tree.tag(), tree.item);
            }
            if (tree instanceof TagTreeAnyItem) {
                return t.thenAnyItem(tree.tag());
            }
            if (tree instanceof TagTreeTag) {
                return t.thenTag(tree.tag());
            }
            return exports.emptyTagTree;
        };
        return dropRec(this, this.depth() - n - 1);
    }
    toNamedString(lookup) {
        const toTagString = (tag) => {
            if (lookup) {
                const keyword = lookup_1.Lookup.keywordOf(tag);
                if (keyword) {
                    return keyword;
                }
            }
            return (0, base_1.tagToString)(tag);
        };
        const toTagTreeString = (tree, tail) => {
            const itemIndexSuffix = tree instanceof TagTreeAnyItem ? '[*]' : 'item' in tree ? '[' + tree.item + ']' : '';
            const head = toTagString(tree.tag()) + itemIndexSuffix;
            const part = head + tail;
            return tree.isRoot() ? part : toTagTreeString(tree.previous(), '.' + part);
        };
        return this.isEmpty() ? '<empty tree>' : toTagTreeString(this, '');
    }
}
exports.TagTree = TagTree;
class TagTreeTrunk extends TagTree {
    thenTag(tag) {
        return new TagTreeTag(tag, this);
    }
    thenAnyItem(tag) {
        return new TagTreeAnyItem(tag, this);
    }
    thenItem(tag, item) {
        return new TagTreeItem(tag, item, this);
    }
}
exports.TagTreeTrunk = TagTreeTrunk;
class EmptyTagTree extends TagTreeTrunk {
    tag() {
        throw Error('Empty tag tree');
    }
    previous() {
        return exports.emptyTagTree;
    }
}
exports.emptyTagTree = new EmptyTagTree(-1, null);
class TagTreeTag extends TagTree {
    constructor(tag, previous) {
        super(tag, previous);
    }
}
exports.TagTreeTag = TagTreeTag;
class TagTreeAnyItem extends TagTreeTrunk {
    constructor(tag, previous) {
        super(tag, previous);
    }
}
exports.TagTreeAnyItem = TagTreeAnyItem;
class TagTreeItem extends TagTreeTrunk {
    constructor(tag, item, previous) {
        super(tag, previous);
        this.item = item;
    }
}
exports.TagTreeItem = TagTreeItem;
//# sourceMappingURL=tag-tree.js.map