"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parsePersonName = exports.Value = void 0;
const js_joda_1 = require("js-joda");
const base_1 = require("./base");
const character_sets_1 = require("./character-sets");
const vr_1 = require("./vr");
const person_name_1 = require("./person-name");
class Value {
    static fromString(vr, value, bigEndian = false) {
        return create(stringBytes(vr, value, bigEndian), vr);
    }
    static fromStrings(vr, values, bigEndian = false) {
        return create(combine(values.map((v) => stringBytes(vr, v, bigEndian)), vr), vr);
    }
    static fromBuffer(vr, buffer) {
        return create(buffer, vr);
    }
    static fromBytes(vr, bytes) {
        return Value.fromBuffer(vr, Buffer.from(bytes));
    }
    static fromNumber(vr, value, bigEndian = false) {
        return create(numberBytes(vr, value, bigEndian), vr);
    }
    static fromNumbers(vr, values, bigEndian = false) {
        return create(combine(values.map((v) => numberBytes(vr, v, bigEndian)), vr), vr);
    }
    static fromDate(vr, value) {
        return create(dateBytes(vr, value), vr);
    }
    static fromDates(vr, values) {
        return create(combine(values.map((v) => dateBytes(vr, v)), vr), vr);
    }
    static fromTime(vr, value) {
        return create(timeBytes(vr, value), vr);
    }
    static fromTimes(vr, values) {
        return create(combine(values.map((v) => timeBytes(vr, v)), vr), vr);
    }
    static fromDateTime(vr, value) {
        return create(dateTimeBytes(vr, value), vr);
    }
    static fromDateTimes(vr, values) {
        return create(combine(values.map((v) => dateTimeBytes(vr, v)), vr), vr);
    }
    static fromPersonName(vr, value) {
        return Value.fromBuffer(vr, personNameBytes(vr, value));
    }
    static fromPersonNames(vr, values) {
        return Value.fromBuffer(vr, combine(values.map((v) => personNameBytes(vr, v)), vr));
    }
    static fromURL(vr, value) {
        return create(urlBytes(vr, value), vr);
    }
    static empty() {
        return new Value(base_1.emptyBuffer);
    }
    static headOption(array) {
        return array.length > 0 ? array[0] : undefined;
    }
    constructor(bytes) {
        this.bytes = bytes;
        this.length = bytes.length;
    }
    toStrings(vr, bigEndian = false, characterSets = character_sets_1.defaultCharacterSet) {
        if (this.length === 0) {
            return [];
        }
        if (vr === vr_1.VR.AT) {
            return parseAT(this.bytes, bigEndian).map(base_1.tagToString);
        }
        if (vr === vr_1.VR.FL) {
            return parseFL(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.FD) {
            return parseFD(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.SS) {
            return parseSS(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.SL) {
            return parseSL(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.SV) {
            return parseSV(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.US) {
            return parseUS(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.UL) {
            return parseUL(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.UV) {
            return parseSV(this.bytes, bigEndian).map((v) => v.toString());
        }
        if (vr === vr_1.VR.OB) {
            return [this.bytes.length + ' bytes'];
        }
        if (vr === vr_1.VR.OW) {
            return [this.bytes.length / 2 + ' words'];
        }
        if (vr === vr_1.VR.OL) {
            return [this.bytes.length / 4 + ' longs'];
        }
        if (vr === vr_1.VR.OV) {
            return [this.bytes.length / 8 + ' very longs'];
        }
        if (vr === vr_1.VR.OF) {
            return [this.bytes.length / 4 + ' floats'];
        }
        if (vr === vr_1.VR.OD) {
            return [this.bytes.length / 8 + ' doubles'];
        }
        if (vr === vr_1.VR.ST || vr === vr_1.VR.LT || vr === vr_1.VR.UT) {
            return [trimTrailing(characterSets.decode(this.bytes, vr), vr.paddingByte)];
        }
        if (vr === vr_1.VR.LO || vr === vr_1.VR.SH || vr === vr_1.VR.UC) {
            return splitString(characterSets.decode(this.bytes, vr));
        }
        if (vr == vr_1.VR.PN) {
            return splitString(characterSets.decode(this.bytes, vr)).map(base_1.trim);
        }
        return splitString(this.bytes.toString()).map(base_1.trim);
    }
    toSingleString(vr, bigEndian = false, characterSets = character_sets_1.defaultCharacterSet) {
        const strings = this.toStrings(vr, bigEndian, characterSets);
        return strings.length === 0 ? '' : strings.join(base_1.multiValueDelimiter);
    }
    toNumbers(vr, bigEndian = false) {
        if (this.length === 0) {
            return [];
        }
        if (vr === vr_1.VR.AT) {
            return parseAT(this.bytes, bigEndian);
        }
        if (vr === vr_1.VR.DS) {
            return parseDS(this.bytes).filter((n) => !isNaN(n));
        }
        if (vr === vr_1.VR.FL) {
            return parseFL(this.bytes, bigEndian);
        }
        if (vr === vr_1.VR.FD) {
            return parseFD(this.bytes, bigEndian);
        }
        if (vr === vr_1.VR.IS) {
            return parseIS(this.bytes).filter((n) => !isNaN(n));
        }
        if (vr === vr_1.VR.SL) {
            return parseSL(this.bytes, bigEndian);
        }
        if (vr === vr_1.VR.SS) {
            return parseSS(this.bytes, bigEndian);
        }
        if (vr === vr_1.VR.UL) {
            return parseUL(this.bytes, bigEndian);
        }
        if (vr === vr_1.VR.US) {
            return parseUS(this.bytes, bigEndian);
        }
        return [];
    }
    toDates(vr = vr_1.VR.DA) {
        if (vr === vr_1.VR.DA) {
            return parseDA(this.bytes);
        }
        if (vr === vr_1.VR.DT) {
            return parseDT(this.bytes, base_1.systemZone).map((dt) => dt.toLocalDate());
        }
        return [];
    }
    toTimes(vr = vr_1.VR.TM) {
        if (vr === vr_1.VR.DT) {
            return parseDT(this.bytes, base_1.systemZone).map((dt) => dt.toLocalTime());
        }
        if (vr === vr_1.VR.TM) {
            return parseTM(this.bytes);
        }
        return [];
    }
    toDateTimes(vr = vr_1.VR.DT, zone = base_1.systemZone) {
        if (vr === vr_1.VR.DA) {
            return parseDA(this.bytes).map((da) => da.atStartOfDay(zone));
        }
        if (vr === vr_1.VR.DT) {
            return parseDT(this.bytes, zone);
        }
        return [];
    }
    toPersonNames(vr = vr_1.VR.PN, characterSets = character_sets_1.defaultCharacterSet) {
        if (vr === vr_1.VR.PN) {
            return parsePN(this.bytes, characterSets);
        }
        return [];
    }
    toURL(vr = vr_1.VR.UR) {
        if (vr === vr_1.VR.UR) {
            return parseUR(this.bytes);
        }
        return undefined;
    }
    toString(vr, bigEndian = false, characterSets = character_sets_1.defaultCharacterSet) {
        return Value.headOption(this.toStrings(vr, bigEndian, characterSets));
    }
    toNumber(vr, bigEndian = false) {
        return Value.headOption(this.toNumbers(vr, bigEndian));
    }
    toDate(vr) {
        return Value.headOption(this.toDates(vr));
    }
    toTime(vr) {
        return Value.headOption(this.toTimes(vr));
    }
    toDateTime(vr, zone) {
        return Value.headOption(this.toDateTimes(vr, zone));
    }
    toPersonName(vr = vr_1.VR.PN, characterSets = character_sets_1.defaultCharacterSet) {
        return Value.headOption(this.toPersonNames(vr, characterSets));
    }
    append(bytes) {
        return new Value((0, base_1.concat)(this.bytes, bytes));
    }
    ensurePadding(vr) {
        return Value.fromBuffer(vr, this.bytes);
    }
}
exports.Value = Value;
function trimTrailing(s, paddingByte) {
    let index = s.length;
    while (index > 0 && s.charCodeAt(index - 1) <= paddingByte) {
        index -= 1;
    }
    return s.substring(0, index);
}
function combine(values, vr) {
    if (vr === vr_1.VR.AT ||
        vr === vr_1.VR.FL ||
        vr === vr_1.VR.FD ||
        vr === vr_1.VR.SL ||
        vr === vr_1.VR.SS ||
        vr === vr_1.VR.UL ||
        vr === vr_1.VR.US ||
        vr === vr_1.VR.OB ||
        vr === vr_1.VR.OW ||
        vr === vr_1.VR.OL ||
        vr === vr_1.VR.OF ||
        vr === vr_1.VR.OD) {
        return values.reduce(base_1.concat, base_1.emptyBuffer);
    }
    const delim = Buffer.from('\\');
    return values.reduce((prev, curr) => (0, base_1.concatv)(prev, delim, curr));
}
function create(bytes, vr) {
    return vr ? new Value((0, base_1.padToEvenLength)(bytes, vr)) : new Value(bytes);
}
function stringBytes(vr, value, bigEndian = false) {
    if (vr === vr_1.VR.AT) {
        return (0, base_1.tagToBytes)(parseInt(value, 16), bigEndian);
    }
    if (vr === vr_1.VR.FL) {
        return (0, base_1.floatToBytes)(parseFloat(value), bigEndian);
    }
    if (vr === vr_1.VR.FD) {
        return (0, base_1.doubleToBytes)(parseFloat(value), bigEndian);
    }
    if (vr === vr_1.VR.SL) {
        return (0, base_1.intToBytes)(parseInt(value, 10), bigEndian);
    }
    if (vr === vr_1.VR.SS) {
        return (0, base_1.shortToBytes)(parseInt(value, 10), bigEndian);
    }
    if (vr === vr_1.VR.UL) {
        return (0, base_1.intToBytes)(parseInt(value, 10), bigEndian);
    }
    if (vr === vr_1.VR.US) {
        return (0, base_1.shortToBytes)(parseInt(value, 10), bigEndian);
    }
    if (vr === vr_1.VR.OB || vr === vr_1.VR.OW || vr === vr_1.VR.OL || vr === vr_1.VR.OF || vr === vr_1.VR.OD) {
        throw Error('Cannot create binary array from string');
    }
    return Buffer.from(value);
}
function numberBytes(vr, value, bigEndian = false) {
    if (vr === vr_1.VR.AT) {
        return (0, base_1.tagToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.FL) {
        return (0, base_1.floatToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.FD) {
        return (0, base_1.doubleToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.SL) {
        return (0, base_1.intToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.SS) {
        return (0, base_1.shortToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.UL) {
        return (0, base_1.intToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.US) {
        return (0, base_1.shortToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.AT) {
        return (0, base_1.tagToBytes)(value, bigEndian);
    }
    if (vr === vr_1.VR.OB || vr === vr_1.VR.OW || vr === vr_1.VR.OL || vr === vr_1.VR.OF || vr === vr_1.VR.OD) {
        throw Error('Cannot create value of VR ' + vr + ' from int');
    }
    return Buffer.from(value + '');
}
function dateBytes(vr, value) {
    if (vr === vr_1.VR.AT ||
        vr === vr_1.VR.FL ||
        vr === vr_1.VR.FD ||
        vr === vr_1.VR.SL ||
        vr === vr_1.VR.SS ||
        vr === vr_1.VR.UL ||
        vr === vr_1.VR.US ||
        vr === vr_1.VR.OB ||
        vr === vr_1.VR.OW ||
        vr === vr_1.VR.OL ||
        vr === vr_1.VR.OF ||
        vr === vr_1.VR.OD) {
        throw Error('Cannot create value of VR ' + vr + ' from date');
    }
    return Buffer.from(formatDate(value));
}
function timeBytes(vr, value) {
    if (vr === vr_1.VR.AT ||
        vr === vr_1.VR.FL ||
        vr === vr_1.VR.FD ||
        vr === vr_1.VR.SL ||
        vr === vr_1.VR.SS ||
        vr === vr_1.VR.UL ||
        vr === vr_1.VR.US ||
        vr === vr_1.VR.OB ||
        vr === vr_1.VR.OW ||
        vr === vr_1.VR.OL ||
        vr === vr_1.VR.OF ||
        vr === vr_1.VR.OD) {
        throw Error('Cannot create value of VR ' + vr + ' from time');
    }
    return Buffer.from(formatTime(value));
}
function dateTimeBytes(vr, value) {
    if (vr === vr_1.VR.AT ||
        vr === vr_1.VR.FL ||
        vr === vr_1.VR.FD ||
        vr === vr_1.VR.SL ||
        vr === vr_1.VR.SS ||
        vr === vr_1.VR.UL ||
        vr === vr_1.VR.US ||
        vr === vr_1.VR.OB ||
        vr === vr_1.VR.OW ||
        vr === vr_1.VR.OL ||
        vr === vr_1.VR.OF ||
        vr === vr_1.VR.OD) {
        throw Error('Cannot create value of VR ' + vr + ' from date-time');
    }
    return Buffer.from(formatDateTime(value));
}
function personNameBytes(vr, value) {
    if (vr === vr_1.VR.PN) {
        return Buffer.from(value.toString());
    }
    throw Error('Cannot create value of VR ' + vr + ' from person name');
}
function urlBytes(vr, value) {
    if (vr === vr_1.VR.UR) {
        return Buffer.from(value.toString());
    }
    throw Error('Cannot create value of VR ' + vr + ' from URL');
}
function chunk(arr, len) {
    const chunks = [];
    const n = arr.length;
    let i = 0;
    while (i < n) {
        chunks.push(arr.slice(i, (i += len)));
    }
    return chunks;
}
function splitFixed(bytes, size) {
    return chunk(bytes, size).filter((g) => g.length === size);
}
function splitString(s) {
    return s.split(base_1.multiValueDelimiter);
}
function parseAT(value, bigEndian = false) {
    return splitFixed(value, 4).map((b) => (0, base_1.bytesToTag)(b, bigEndian));
}
function parseSL(value, bigEndian = false) {
    return splitFixed(value, 4).map((b) => (0, base_1.bytesToInt)(b, bigEndian));
}
function parseSV(value, bigEndian = false) {
    return splitFixed(value, 8).map((b) => (0, base_1.bytesToFloat)(b, bigEndian));
}
function parseSS(value, bigEndian = false) {
    return splitFixed(value, 2).map((b) => (0, base_1.bytesToShort)(b, bigEndian));
}
function parseUL(value, bigEndian = false) {
    return splitFixed(value, 4).map((b) => (0, base_1.bytesToUInt)(b, bigEndian));
}
function parseUS(value, bigEndian = false) {
    return splitFixed(value, 2).map((b) => (0, base_1.bytesToUShort)(b, bigEndian));
}
function parseFL(value, bigEndian = false) {
    return splitFixed(value, 4).map((b) => (0, base_1.bytesToFloat)(b, bigEndian));
}
function parseFD(value, bigEndian = false) {
    return splitFixed(value, 8).map((b) => (0, base_1.bytesToDouble)(b, bigEndian));
}
function parseDS(value) {
    return splitString(value.toString())
        .map(base_1.trim)
        .map((s) => parseFloat(s));
}
function parseIS(value) {
    return splitString(value.toString())
        .map(base_1.trim)
        .map((s) => parseInt(s, 10));
}
function parseDA(value) {
    return splitString(value.toString())
        .map(parseDate)
        .filter((d) => d !== undefined);
}
function parseTM(value) {
    return splitString(value.toString())
        .map(parseTime)
        .filter((d) => d !== undefined);
}
function parseDT(value, zone) {
    return splitString(value.toString())
        .map((s) => parseDateTime(s, zone))
        .filter((d) => d !== undefined);
}
function parsePN(value, characterSets) {
    return splitString(characterSets.decode(value, vr_1.VR.PN))
        .map(base_1.trim)
        .map((s) => parsePersonName(s));
}
function parseUR(value) {
    return parseURL(value.toString().trim());
}
const dateFormat1 = js_joda_1.DateTimeFormatter.ofPattern('uuuuMMdd');
const dateFormat2 = js_joda_1.DateTimeFormatter.ofPattern("uuuu'.'MM'.'dd");
const timeFormat = new js_joda_1.DateTimeFormatterBuilder()
    .appendPattern("HH[[':']mm[[':']ss[")
    .appendFraction(js_joda_1.ChronoField.MICRO_OF_SECOND, 1, 6, true)
    .appendPattern(']]]')
    .toFormatter(js_joda_1.ResolverStyle.LENIENT);
const timeFormatForEncoding = js_joda_1.DateTimeFormatter.ofPattern("HHmmss'.'SSSSSS");
const dateTimeFormatForEncoding = js_joda_1.DateTimeFormatter.ofPattern("uuuuMMddHHmmss'.'SSSSSSZ");
function formatDate(date) {
    return date.format(dateFormat1);
}
function formatTime(time) {
    return time.format(timeFormatForEncoding);
}
function formatDateTime(dateTime) {
    return dateTime.format(dateTimeFormatForEncoding);
}
function parseDate(s) {
    const trimmed = s.trim();
    try {
        return js_joda_1.LocalDate.parse(trimmed, dateFormat1);
    }
    catch (error) {
        try {
            return js_joda_1.LocalDate.parse(trimmed, dateFormat2);
        }
        catch (error) {
            return undefined;
        }
    }
}
function parseTime(s) {
    try {
        return js_joda_1.LocalTime.parse(s.trim(), timeFormat);
    }
    catch (error) {
        return undefined;
    }
}
function parseDateTime(s, zone = base_1.systemZone) {
    try {
        s = s.trim();
        let len = s.length;
        if (len < 4) {
            throw Error('Malformed date-time, must at least include year.');
        }
        // parse zone if present and trim this part from string
        const zoneStart = Math.max(s.indexOf('+'), s.indexOf('-'));
        if (zoneStart >= 4 && s.length === zoneStart + 5) {
            const signString = s.substring(zoneStart, zoneStart + 1);
            const zoneString = s.substring(zoneStart + 1, zoneStart + 5);
            zone = js_joda_1.ZoneOffset.ofHoursMinutes(parseInt(signString + zoneString.substring(0, 2), 10), parseInt(signString + zoneString.substring(2, 4), 10));
            s = s.substring(0, zoneStart);
            len = s.length;
        }
        else if (zoneStart >= 0) {
            throw Error('Malformed date-time. Zone is present but misplaced or not of length 5');
        }
        const validLengths = [4, 6, 8, 10, 12, 14, 16, 17, 18, 19, 20, 21];
        if (validLengths.indexOf(len) < 0) {
            throw Error('Malformed date-time, invalid length ' + len);
        }
        const year = parseInt(s.substring(0, 4), 10);
        let month = 1;
        let dayOfMonth = 1;
        let hour = 0;
        let minute = 0;
        let second = 0;
        let nanoOfSecond = 0;
        if (len >= 6) {
            month = parseInt(s.substring(4, 6), 10);
            if (len >= 8) {
                dayOfMonth = parseInt(s.substring(6, 8), 10);
                if (len >= 10) {
                    hour = parseInt(s.substring(8, 10), 10);
                    if (len >= 12) {
                        minute = parseInt(s.substring(10, 12), 10);
                        if (len >= 14) {
                            second = parseInt(s.substring(12, 14), 10);
                            if (s.charAt(14) === '.' && len >= 15) {
                                const end = Math.min(len, 21);
                                const precision = end - 15;
                                const exponent = 9 - precision;
                                nanoOfSecond = parseInt(s.substring(15, end), 10) * Math.pow(10, exponent);
                            }
                        }
                    }
                }
            }
        }
        return js_joda_1.ZonedDateTime.of(year, month, dayOfMonth, hour, minute, second, nanoOfSecond, zone);
    }
    catch (error) {
        return undefined;
    }
}
function parsePersonName(s) {
    function ensureLength(ss, n) {
        return ss.concat(new Array(Math.max(0, n - ss.length)).fill(''));
    }
    function transpose(matrix) {
        return matrix[0].map((_, i) => matrix.map((col) => col[i]));
    }
    const matrix = ensureLength(s.split(/=/), 3)
        .map(base_1.trim)
        .map((s1) => ensureLength(s1.split(/\^/), 5));
    const comps = transpose(matrix).map((c) => new person_name_1.ComponentGroup(c[0], c[1], c[2]));
    return new person_name_1.PersonName(comps[0], comps[1], comps[2], comps[3], comps[4]);
}
exports.parsePersonName = parsePersonName;
function parseURL(s) {
    try {
        return new URL(s);
    }
    catch (error) {
        return undefined;
    }
}
//# sourceMappingURL=value.js.map